from dataclasses import dataclass
from decimal import Decimal

from eth_abi.abi import encode
from hexbytes import HexBytes
from web3 import Account, Web3

from hummingbot.connector.derivative.derive_perpetual.derive_perpetual_web_utils import decimal_to_big_int


@dataclass
class ModuleData:
    def to_abi_encoded(self):
        pass

    def to_json(self):
        pass


@dataclass
class TradeModuleData(ModuleData):
    asset_address: str
    sub_id: int
    limit_price: Decimal
    amount: Decimal
    max_fee: Decimal
    recipient_id: int
    is_bid: bool

    def to_abi_encoded(self):
        return encode(
            ["address", "uint", "int", "int", "uint", "uint", "bool"],
            [
                Web3.to_checksum_address(self.asset_address),
                self.sub_id,
                decimal_to_big_int(self.limit_price),
                decimal_to_big_int(self.amount),
                decimal_to_big_int(self.max_fee),
                self.recipient_id,
                self.is_bid,
            ],
        )

    def to_json(self):
        return {
            "limit_price": str(self.limit_price),
            "amount": str(self.amount),
            "max_fee": str(self.max_fee),
        }


@dataclass
class SignedAction:
    subaccount_id: int
    owner: str
    signer: str
    signature_expiry_sec: int
    nonce: int
    module_address: str
    module_data: ModuleData
    DOMAIN_SEPARATOR: str
    ACTION_TYPEHASH: str
    signature: str = ""
    """
    Used to sign and validate actions.

    :param subaccount_id: The subaccount id of the user.
    :param owner: The owner of the account on the v2 protocol (not the session key).
    :param signer: The signer of the action - can be the owner or a session key.
    :param signature_expiry_sec: The expiry time of the signature in seconds. Must be >5min from now.
    :param nonce: Unique nonce defined as <UTC_timestamp in ms><random_number_up_to_6_digits> (e.g. 1695836058725001, where 001 is the random number).
    :param module_address: The contract address of the module. Refer to Protocol Constants table in docs.derive.xyz.
    :param module_data: Data defined by the specific protocol module (e.g. for orders use module_data.trade.TradeModuleData).
    :param DOMAIN_SEPARATOR: The domain separator of the protocol. Refer to Protocol Constants table in docs.derive.xyz.
    :param ACTION_TYPEHASH: The typehash of the action. Refer to Protocol Constants table in docs.derive.xyz.
    :param signature: The signature of the action. Use sign() to generate the signature.
    """

    def sign(self, signer_private_key: str):
        signer_wallet = Web3().eth.account.from_key(signer_private_key)
        signature: Account = signer_wallet.unsafe_sign_hash(self._to_typed_data_hash())
        self.signature = signature.signature.hex()
        return self.signature

    def to_json(self):
        return {
            "subaccount_id": self.subaccount_id,
            "nonce": self.nonce,
            "signer": self.signer,
            "signature_expiry_sec": self.signature_expiry_sec,
            "signature": self.signature,
            **self.module_data.to_json(),
        }

    def validate_signature(self):
        data_hash = self._to_typed_data_hash()
        recovered = Account._recover_hash(
            data_hash.hex(),
            signature=HexBytes(self.signature),
        )

        if recovered.lower() != self.signer.lower():
            raise ValueError("Invalid signature. Recovered signer does not match expected signer.")

    @property
    def domain_separator(self) -> bytes:
        try:
            return bytes.fromhex(self.DOMAIN_SEPARATOR[2:])
        except ValueError:
            raise ValueError(
                "Unable to extract bytes from DOMAIN_SEPARATOR. Ensure value is copied from Protocol Constants in docs.derive.xyz."
            )

    @property
    def action_typehash(self) -> bytes:
        try:
            return bytes.fromhex(self.ACTION_TYPEHASH[2:])
        except ValueError:
            raise ValueError(
                "Unable to extract bytes from ACTION_TYPEHASH. Ensure value is copied from Protocol Constants in docs.derive.xyz."
            )

    def _to_typed_data_hash(self) -> HexBytes:
        encoded_typed_data_hash = "".join(["0x1901", self.DOMAIN_SEPARATOR[2:], self._get_action_hash().hex()])
        return Web3.keccak(hexstr=encoded_typed_data_hash)

    def _get_action_hash(self) -> HexBytes:
        return Web3.keccak(
            encode(
                [
                    "bytes32",
                    "uint",
                    "uint",
                    "address",
                    "bytes32",
                    "uint",
                    "address",
                    "address",
                ],
                [
                    self.action_typehash,
                    self.subaccount_id,
                    self.nonce,
                    Web3.to_checksum_address(self.module_address),
                    Web3.keccak(self.module_data.to_abi_encoded()),
                    self.signature_expiry_sec,
                    Web3.to_checksum_address(self.owner),
                    Web3.to_checksum_address(self.signer),
                ],
            )
        )



from hummingbot.core.api_throttler.data_types import LinkedLimitWeightPair, RateLimit
from hummingbot.core.data_type.common import OrderType
from hummingbot.core.data_type.in_flight_order import OrderState

# A single source of truth for constant variables related to the exchange

EXCHANGE_NAME = "dydx_v4_perpetual"
DEFAULT_DOMAIN = "com"

API_VERSION = "v4"
CURRENCY = "USD"

HBOT_BROKER_ID = "Hummingbot"
MAX_ID_LEN = 40
HEARTBEAT_INTERVAL = 30.0
ORDER_EXPIRATION = 2419200  # 28 days

# height limited to 2147483647 for 32-bit OS, equivalent to 2038-01-19T03:14Z
TX_MAX_HEIGHT = 2147483647
LIMIT_FEE = 0.015

# API Base URLs
MAX_ID_BIT_COUNT = 31

# data_source grpc
DYDX_V4_AERIAL_GRPC_OR_REST_PREFIX = "grpc"
DYDX_V4_AERIAL_CONFIG_URL = 'dydx-grpc.publicnode.com:443'
DYDX_V4_QUERY_AERIAL_CONFIG_URL = 'dydx-grpc.publicnode.com:443'
CHAIN_ID = 'dydx-mainnet-1'
FEE_DENOMINATION = "afet"
TX_FEE = 0
TX_GAS_LIMIT = 0

DYDX_V4_VALIDATOR_REST_BASE_URL = "https://dydx-grpc.publicnode.com:443"

DYDX_V4_INDEXER_REST_BASE_URL = "https://indexer.dydx.trade"

DYDX_V4_REST_URL = "{}/{}".format(DYDX_V4_INDEXER_REST_BASE_URL, API_VERSION)

DYDX_V4_WS_URL = "wss://indexer.dydx.trade/{}/ws".format(API_VERSION)

# Public REST Endpoints

PATH_MARKETS = "/perpetualMarkets"

PATH_HISTORY_FUNDING = "/historicalFunding"
PATH_TICKER = "/stats"

PATH_SNAPSHOT = "/orderbooks/perpetualMarket"
PATH_TIME = "/time"

PATH_ORDERS = "/orders"

PATH_FILLS = "/fills"
PATH_POSITIONS = "/perpetualPositions"

PATH_ACCOUNTS = "/accounts"
PATH_CONFIG = "/config"

PATH_FUNDING = "/historical-pnl"

PATH_SUBACCOUNT = "/addresses"

# WS Endpoints
WS_PATH_ACCOUNTS = "/ws/accounts"

# WS Channels

WS_CHANNEL_TRADES = "v4_trades"
WS_CHANNEL_ORDERBOOK = "v4_orderbook"
WS_CHANNEL_MARKETS = "v4_markets"
WS_CHANNEL_ACCOUNTS = "v4_subaccounts"

WS_TYPE_SUBSCRIBE = "subscribe"
WS_TYPE_SUBSCRIBED = "subscribed"
WS_TYPE_CHANNEL_DATA = "channel_data"

TIF_GOOD_TIL_TIME = "GTT"
TIF_FILL_OR_KILL = "FOK"
TIF_IMMEDIATE_OR_CANCEL = "IOC"

FEES_KEY = "*"
FEE_MAKER_KEY = "maker"
FEE_TAKER_KEY = "taker"

ORDER_TYPE_MAP = {
    OrderType.LIMIT: "LIMIT",
    OrderType.LIMIT_MAKER: "LIMIT",
    OrderType.MARKET: "MARKET",
}

ORDER_STATE = {
    "PENDING": OrderState.OPEN,
    "OPEN": OrderState.OPEN,
    "BEST_EFFORT_OPENED": OrderState.OPEN,
    "FILLED": OrderState.FILLED,
    "CANCELED": OrderState.CANCELED,
    "BEST_EFFORT_CANCELED": OrderState.PENDING_CANCEL,
}

WS_CHANNEL_TO_PATH = {WS_CHANNEL_ACCOUNTS: WS_PATH_ACCOUNTS}

LAST_FEE_PAYMENTS_MAX = 1
LAST_FILLS_MAX = 100

LIMIT_ID_GET = "LIMIT_ID_GET"
LIMIT_ID_ORDER_CANCEL = "LIMIT_ID_ORDER_CANCEL"
LIMIT_ID_LONG_TERM_ORDER_PLACE = "LIMIT_ID_LONG_TERM_ORDER_PLACE"

LIMIT_LONG_TERM_ORDER_PLACE = "LIMIT_LONG_TERM_ORDER_PLACE"
MARKET_SHORT_TERM_ORDER_PLACE = "LIMIT_LONG_TERM_ORDER_PLACE"

NO_LIMIT = 1000
ONE_SECOND = 1
ONE_HUNDRED_SECOND = 100

QUOTE_QUANTUMS_ATOMIC_RESOLUTION = -6
ORDER_FLAGS_SHORT_TERM = 0
ORDER_FLAGS_LONG_TERM = 64

TIME_IN_FORCE_IOC = 1
TIME_IN_FORCE_POST_ONLY = 2
TIME_IN_FORCE_UNSPECIFIED = 0

RATE_LIMITS = [
    # Pools
    RateLimit(limit_id=LIMIT_ID_GET, limit=NO_LIMIT, time_interval=ONE_SECOND),
    RateLimit(limit_id=LIMIT_LONG_TERM_ORDER_PLACE, limit=20, time_interval=ONE_HUNDRED_SECOND),
    RateLimit(limit_id=MARKET_SHORT_TERM_ORDER_PLACE, limit=200, time_interval=ONE_SECOND),
    # Weighted limits
    RateLimit(
        limit_id=PATH_CONFIG,
        limit=NO_LIMIT,
        time_interval=ONE_SECOND,
        linked_limits=[LinkedLimitWeightPair(LIMIT_ID_GET)],
    ),
    RateLimit(
        limit_id=PATH_FILLS,
        limit=NO_LIMIT,
        time_interval=ONE_SECOND,
        linked_limits=[LinkedLimitWeightPair(LIMIT_ID_GET)],
    ),
    RateLimit(
        limit_id=PATH_ORDERS,
        limit=NO_LIMIT,
        time_interval=ONE_SECOND,
        linked_limits=[LinkedLimitWeightPair(LIMIT_ID_GET)],
    ),
    RateLimit(
        limit_id=PATH_FUNDING,
        limit=NO_LIMIT,
        time_interval=ONE_SECOND,
        linked_limits=[LinkedLimitWeightPair(LIMIT_ID_GET)],
    ),
    RateLimit(
        limit_id=PATH_HISTORY_FUNDING,
        limit=NO_LIMIT,
        time_interval=ONE_SECOND,
        linked_limits=[LinkedLimitWeightPair(LIMIT_ID_GET)],
    ),
    RateLimit(
        limit_id=PATH_ACCOUNTS,
        limit=NO_LIMIT,
        time_interval=ONE_SECOND,
        linked_limits=[LinkedLimitWeightPair(LIMIT_ID_GET)],
    ),
    RateLimit(
        limit_id=PATH_SUBACCOUNT,
        limit=NO_LIMIT,
        time_interval=ONE_SECOND,
        linked_limits=[LinkedLimitWeightPair(LIMIT_ID_GET)],
    ),
    RateLimit(
        limit_id=PATH_MARKETS,
        limit=NO_LIMIT,
        time_interval=ONE_SECOND,
        linked_limits=[LinkedLimitWeightPair(LIMIT_ID_GET)],
    ),
    RateLimit(
        limit_id=PATH_TIME,
        limit=NO_LIMIT,
        time_interval=ONE_SECOND,
        linked_limits=[LinkedLimitWeightPair(LIMIT_ID_GET)],
    ),
    RateLimit(
        limit_id=PATH_SNAPSHOT,
        limit=NO_LIMIT,
        time_interval=ONE_SECOND,
        linked_limits=[LinkedLimitWeightPair(LIMIT_ID_GET)],
    ),
    RateLimit(
        limit_id=LIMIT_ID_LONG_TERM_ORDER_PLACE,
        limit=2,
        time_interval=ONE_SECOND,
        linked_limits=[LinkedLimitWeightPair(LIMIT_LONG_TERM_ORDER_PLACE)],
    ),
    RateLimit(
        limit_id=LIMIT_ID_ORDER_CANCEL,
        limit=NO_LIMIT,
        time_interval=ONE_SECOND,

    ),
]

ACCOUNT_SEQUENCE_MISMATCH_ERROR = "account sequence mismatch"
ERR_MSG_NO_ORDER_FOUND = "Stateful order does not exist"



from decimal import Decimal
from typing import List, Tuple

from hummingbot.connector.exchange.paper_trade import create_paper_trade_market
from hummingbot.connector.exchange_base import ExchangeBase
from hummingbot.strategy.api_asset_price_delegate import APIAssetPriceDelegate
from hummingbot.strategy.market_trading_pair_tuple import MarketTradingPairTuple
from hummingbot.strategy.order_book_asset_price_delegate import OrderBookAssetPriceDelegate
from hummingbot.strategy.perpetual_market_making import PerpetualMarketMakingStrategy
from hummingbot.strategy.perpetual_market_making.perpetual_market_making_config_map import (
    perpetual_market_making_config_map as c_map,
)


async def start(self):
    try:
        leverage = c_map.get("leverage").value
        position_mode = c_map.get("position_mode").value
        order_amount = c_map.get("order_amount").value
        order_refresh_time = c_map.get("order_refresh_time").value
        bid_spread = c_map.get("bid_spread").value / Decimal('100')
        ask_spread = c_map.get("ask_spread").value / Decimal('100')
        long_profit_taking_spread = c_map.get("long_profit_taking_spread").value / Decimal('100')
        short_profit_taking_spread = c_map.get("short_profit_taking_spread").value / Decimal('100')
        stop_loss_spread = c_map.get("stop_loss_spread").value / Decimal('100')
        time_between_stop_loss_orders = c_map.get("time_between_stop_loss_orders").value
        stop_loss_slippage_buffer = c_map.get("stop_loss_slippage_buffer").value / Decimal('100')
        minimum_spread = c_map.get("minimum_spread").value / Decimal('100')
        price_ceiling = c_map.get("price_ceiling").value
        price_floor = c_map.get("price_floor").value
        order_levels = c_map.get("order_levels").value
        order_level_amount = c_map.get("order_level_amount").value
        order_level_spread = c_map.get("order_level_spread").value / Decimal('100')
        exchange = c_map.get("derivative").value.lower()
        raw_trading_pair = c_map.get("market").value
        filled_order_delay = c_map.get("filled_order_delay").value
        order_optimization_enabled = c_map.get("order_optimization_enabled").value
        ask_order_optimization_depth = c_map.get("ask_order_optimization_depth").value
        bid_order_optimization_depth = c_map.get("bid_order_optimization_depth").value
        price_source = c_map.get("price_source").value
        price_type = c_map.get("price_type").value
        price_source_exchange = c_map.get("price_source_derivative").value
        price_source_market = c_map.get("price_source_market").value
        price_source_custom_api = c_map.get("price_source_custom_api").value
        custom_api_update_interval = c_map.get("custom_api_update_interval").value
        order_refresh_tolerance_pct = c_map.get("order_refresh_tolerance_pct").value / Decimal('100')
        order_override = c_map.get("order_override").value

        trading_pair: str = raw_trading_pair
        base, quote = trading_pair.split("-")
        maker_assets: Tuple[str, str] = (base, quote)
        market_names: List[Tuple[str, List[str]]] = [(exchange, [trading_pair])]
        await self.initialize_markets(market_names)
        maker_data = [self.markets[exchange], trading_pair] + list(maker_assets)
        self.market_trading_pair_tuples = [MarketTradingPairTuple(*maker_data)]
        asset_price_delegate = None
        if price_source == "external_market":
            asset_trading_pair: str = price_source_market
            ext_market = create_paper_trade_market(
                price_source_exchange, [asset_trading_pair]
            )
            self.markets[price_source_exchange]: ExchangeBase = ext_market
            asset_price_delegate = OrderBookAssetPriceDelegate(ext_market, asset_trading_pair)
        elif price_source == "custom_api":
            ext_market = create_paper_trade_market(
                exchange, [raw_trading_pair]
            )
            asset_price_delegate = APIAssetPriceDelegate(ext_market, price_source_custom_api,
                                                         custom_api_update_interval)

        strategy_logging_options = PerpetualMarketMakingStrategy.OPTION_LOG_ALL

        self.strategy = PerpetualMarketMakingStrategy()
        self.strategy.init_params(
            market_info=MarketTradingPairTuple(*maker_data),
            leverage=leverage,
            position_mode=position_mode,
            bid_spread=bid_spread,
            ask_spread=ask_spread,
            order_amount=order_amount,
            long_profit_taking_spread=long_profit_taking_spread,
            short_profit_taking_spread=short_profit_taking_spread,
            stop_loss_spread=stop_loss_spread,
            time_between_stop_loss_orders=time_between_stop_loss_orders,
            stop_loss_slippage_buffer=stop_loss_slippage_buffer,
            order_levels=order_levels,
            order_level_spread=order_level_spread,
            order_level_amount=order_level_amount,
            order_refresh_time=order_refresh_time,
            order_refresh_tolerance_pct=order_refresh_tolerance_pct,
            filled_order_delay=filled_order_delay,
            order_optimization_enabled=order_optimization_enabled,
            ask_order_optimization_depth=ask_order_optimization_depth,
            bid_order_optimization_depth=bid_order_optimization_depth,
            asset_price_delegate=asset_price_delegate,
            price_type=price_type,
            price_ceiling=price_ceiling,
            price_floor=price_floor,
            logging_options=strategy_logging_options,
            minimum_spread=minimum_spread,
            hb_app_notification=True,
            order_override=order_override,
        )
    except Exception as e:
        self.notify(str(e))
        self.logger().error("Unknown error during initialization.", exc_info=True)



# distutils: language=c++

from hummingbot.strategy.order_tracker import OrderTracker
from hummingbot.strategy.order_tracker cimport OrderTracker


cdef class PureMarketMakingOrderTracker(OrderTracker):
    pass



from decimal import Decimal
from typing import Literal, Optional

from pydantic import ConfigDict, Field, SecretStr, field_validator

from hummingbot.client.config.config_data_types import BaseConnectorConfigMap
from hummingbot.core.data_type.trade_fee import TradeFeeSchema

# Maker rebates(-0.02%) are paid out continuously on each trade directly to the trading wallet.(https://hyperliquid.gitbook.io/hyperliquid-docs/trading/fees)
DEFAULT_FEES = TradeFeeSchema(
    maker_percent_fee_decimal=Decimal("0"),
    taker_percent_fee_decimal=Decimal("0.00025"),
    buy_percent_fee_deducted_from_returns=True
)

CENTRALIZED = True

EXAMPLE_PAIR = "BTC-USD"

BROKER_ID = "HBOT"


def validate_wallet_mode(value: str) -> Optional[str]:
    """
    Check if the value is a valid mode
    """
    allowed = ('arb_wallet', 'api_wallet')

    if isinstance(value, str):
        formatted_value = value.strip().lower()

        if formatted_value in allowed:
            return formatted_value

    raise ValueError(f"Invalid wallet mode '{value}', choose from: {allowed}")


def validate_bool(value: str) -> Optional[str]:
    """
    Permissively interpret a string as a boolean
    """
    if isinstance(value, bool):
        return value

    if isinstance(value, str):
        formatted_value = value.strip().lower()
        truthy = {"yes", "y", "true", "1"}
        falsy = {"no", "n", "false", "0"}

        if formatted_value in truthy:
            return True
        if formatted_value in falsy:
            return False

    raise ValueError(f"Invalid value, please choose value from {truthy.union(falsy)}")


class HyperliquidPerpetualConfigMap(BaseConnectorConfigMap):
    connector: str = "hyperliquid_perpetual"
    hyperliquid_perpetual_mode: Literal["arb_wallet", "api_wallet"] = Field(
        default="arb_wallet",
        json_schema_extra={
            "prompt": "Select connection mode (arb_wallet/api_wallet)",
            "is_secure": False,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    use_vault: bool = Field(
        default="no",
        json_schema_extra={
            "prompt": "Do you want to use the Vault address? (Yes/No)",
            "is_secure": False,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    hyperliquid_perpetual_address: SecretStr = Field(
        default=...,
        json_schema_extra={
            "prompt": lambda cm: (
                "Enter your Vault address"
                if getattr(cm, "use_vault", False)
                else "Enter your Arbitrum wallet address"
            ),
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    hyperliquid_perpetual_secret_key: SecretStr = Field(
        default=...,
        json_schema_extra={
            "prompt": lambda cm: {
                "arb_wallet": "Enter your Arbitrum wallet private key",
                "api_wallet": "Enter your API wallet private key (from https://app.hyperliquid.xyz/API)"
            }.get(getattr(cm, "hyperliquid_perpetual_mode", "arb_wallet")),
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    model_config = ConfigDict(title="hyperliquid_perpetual")

    @field_validator("hyperliquid_perpetual_mode", mode="before")
    @classmethod
    def validate_mode(cls, value: str) -> str:
        """Used for client-friendly error output."""
        return validate_wallet_mode(value)

    @field_validator("use_vault", mode="before")
    @classmethod
    def validate_use_vault(cls, value: str):
        """Used for client-friendly error output."""
        return validate_bool(value)

    @field_validator("hyperliquid_perpetual_address", mode="before")
    @classmethod
    def validate_address(cls, value: str):
        """Used for client-friendly error output."""
        if isinstance(value, str):
            if value.startswith("HL:"):
                # Strip out the "HL:" that the HyperLiquid Vault page adds to vault addresses
                return value[3:]
        return value


KEYS = HyperliquidPerpetualConfigMap.model_construct()

OTHER_DOMAINS = ["hyperliquid_perpetual_testnet"]
OTHER_DOMAINS_PARAMETER = {"hyperliquid_perpetual_testnet": "hyperliquid_perpetual_testnet"}
OTHER_DOMAINS_EXAMPLE_PAIR = {"hyperliquid_perpetual_testnet": "BTC-USD"}
OTHER_DOMAINS_DEFAULT_FEES = {"hyperliquid_perpetual_testnet": [0, 0.025]}


class HyperliquidPerpetualTestnetConfigMap(BaseConnectorConfigMap):
    connector: str = "hyperliquid_perpetual_testnet"
    hyperliquid_perpetual_testnet_mode: Literal["arb_wallet", "api_wallet"] = Field(
        default="arb_wallet",
        json_schema_extra={
            "prompt": "Select connection mode (arb_wallet/api_wallet)",
            "is_secure": False,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    use_vault: bool = Field(
        default="no",
        json_schema_extra={
            "prompt": "Do you want to use the Vault address? (Yes/No)",
            "is_secure": False,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    hyperliquid_perpetual_testnet_address: SecretStr = Field(
        default=...,
        json_schema_extra={
            "prompt": lambda cm: (
                "Enter your Vault address"
                if getattr(cm, "use_vault", False)
                else "Enter your Arbitrum wallet address"
            ),
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    hyperliquid_perpetual_testnet_secret_key: SecretStr = Field(
        default=...,
        json_schema_extra={
            "prompt": lambda cm: {
                "arb_wallet": "Enter your Arbitrum wallet private key",
                "api_wallet": "Enter your API wallet private key (from https://app.hyperliquid.xyz/API)"
            }.get(getattr(cm, "hyperliquid_perpetual_testnet_mode", "arb_wallet")),
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    model_config = ConfigDict(title="hyperliquid_perpetual")

    @field_validator("hyperliquid_perpetual_testnet_mode", mode="before")
    @classmethod
    def validate_mode(cls, value: str) -> str:
        """Used for client-friendly error output."""
        return validate_wallet_mode(value)

    @field_validator("use_vault", mode="before")
    @classmethod
    def validate_use_vault(cls, value: str):
        """Used for client-friendly error output."""
        return validate_bool(value)

    @field_validator("hyperliquid_perpetual_testnet_address", mode="before")
    @classmethod
    def validate_address(cls, value: str):
        """Used for client-friendly error output."""
        if isinstance(value, str):
            if value.startswith("HL:"):
                # Strip out the "HL:" that the HyperLiquid Vault page adds to vault addresses
                return value[3:]
        return value


OTHER_DOMAINS_KEYS = {
    "hyperliquid_perpetual_testnet": HyperliquidPerpetualTestnetConfigMap.model_construct()
}



from decimal import Decimal

from sqlalchemy import JSON, BigInteger, Boolean, Column, Float, Index, Integer, Text

from hummingbot.model import HummingbotBase
from hummingbot.strategy_v2.models.base import RunnableStatus
from hummingbot.strategy_v2.models.executors import CloseType
from hummingbot.strategy_v2.models.executors_info import ExecutorInfo


class Executors(HummingbotBase):
    __tablename__ = "Executors"
    __table_args__ = (
        Index("ex_type", "type"),
        Index("ex_type_timestamp", "type", "timestamp"),
        Index("ex_timestamp", "timestamp"),
        Index("ex_close_timestamp", "close_timestamp"),
        Index("ex_status", "status"),
        Index("ex_type_status", "type", "status"),
    )
    id = Column(Text, primary_key=True)
    timestamp = Column(Float, nullable=False)
    type = Column(Text, nullable=False)
    close_type = Column(Integer, nullable=True)
    close_timestamp = Column(BigInteger, nullable=True)
    status = Column(Integer, nullable=False)
    config = Column(JSON, nullable=False)
    net_pnl_pct = Column(Float, nullable=False)
    net_pnl_quote = Column(Float, nullable=False)
    cum_fees_quote = Column(Float, nullable=False)
    filled_amount_quote = Column(Float, nullable=False)
    is_active = Column(Boolean, nullable=False)
    is_trading = Column(Boolean, nullable=False)
    custom_info = Column(JSON, nullable=False)
    controller_id = Column(Text, nullable=True)

    def to_executor_info(self) -> ExecutorInfo:
        """
        Return an ExecutorInfo object based on the current instance.
        """
        close_type = CloseType(self.close_type) if self.close_type else None
        status = RunnableStatus(self.status)
        return ExecutorInfo(
            id=self.id,
            timestamp=self.timestamp,
            type=self.type,
            close_type=close_type,
            close_timestamp=self.close_timestamp,
            status=status,
            config=self.config,
            net_pnl_pct=Decimal(self.net_pnl_pct),
            net_pnl_quote=Decimal(self.net_pnl_quote),
            cum_fees_quote=Decimal(self.cum_fees_quote),
            filled_amount_quote=Decimal(self.filled_amount_quote),
            is_active=self.is_active,
            is_trading=self.is_trading,
            custom_info=self.custom_info,
            controller_id=self.controller_id,
        )



from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, Optional, TypedDict


class Chain(Enum):
    ETHEREUM = ('ethereum', 'ETH')
    SOLANA = ('solana', 'SOL')

    def __init__(self, chain: str, native_currency: str):
        self.chain = chain
        self.native_currency = native_currency


class Connector(Enum):
    def __int__(self, chain: Chain, connector: str):
        self.chain = chain
        self.connector = connector


class ConnectorType(Enum):
    SWAP = "SWAP"
    CLMM = "CLMM"
    AMM = "AMM"


class TransactionStatus(Enum):
    """Transaction status constants for gateway operations."""
    CONFIRMED = 1
    PENDING = 0
    FAILED = -1


class Token(TypedDict):
    """Token information from gateway."""
    symbol: str
    address: str
    decimals: int
    name: str


def get_connector_type(connector_name: str) -> ConnectorType:
    if "/clmm" in connector_name:
        return ConnectorType.CLMM
    elif "/amm" in connector_name:
        return ConnectorType.AMM
    return ConnectorType.SWAP


@dataclass
class PlaceOrderResult:
    update_timestamp: float
    client_order_id: str
    exchange_order_id: Optional[str]
    trading_pair: str
    misc_updates: Dict[str, Any] = field(default_factory=lambda: {})
    exception: Optional[Exception] = None


@dataclass
class CancelOrderResult:
    client_order_id: str
    trading_pair: str
    misc_updates: Dict[str, Any] = field(default_factory=lambda: {})
    not_found: bool = False
    exception: Optional[Exception] = None



import asyncio
import sys
import time
from decimal import Decimal
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

import dateutil.parser as dp

from hummingbot.connector.derivative.dydx_v4_perpetual import (
    dydx_v4_perpetual_constants as CONSTANTS,
    dydx_v4_perpetual_web_utils as web_utils,
)
from hummingbot.core.data_type.common import TradeType
from hummingbot.core.data_type.funding_info import FundingInfo, FundingInfoUpdate
from hummingbot.core.data_type.order_book import OrderBookMessage
from hummingbot.core.data_type.order_book_message import OrderBookMessageType
from hummingbot.core.data_type.perpetual_api_order_book_data_source import PerpetualAPIOrderBookDataSource
from hummingbot.core.utils.tracking_nonce import NonceCreator
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, WSJSONRequest
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant

if TYPE_CHECKING:
    from hummingbot.connector.derivative.dydx_v4_perpetual.dydx_v4_perpetual_derivative import DydxV4PerpetualDerivative


class DydxV4PerpetualAPIOrderBookDataSource(PerpetualAPIOrderBookDataSource):
    FULL_ORDER_BOOK_RESET_DELTA_SECONDS = sys.maxsize

    _DYNAMIC_SUBSCRIBE_ID_START = 100
    _next_subscribe_id: int = _DYNAMIC_SUBSCRIBE_ID_START

    def __init__(
            self,
            trading_pairs: List[str],
            connector: "DydxV4PerpetualDerivative",
            api_factory: WebAssistantsFactory,
            domain: str = CONSTANTS.DEFAULT_DOMAIN,
    ):
        super().__init__(trading_pairs)
        self._connector = connector
        self._api_factory = api_factory
        self._domain = domain
        self._nonce_provider = NonceCreator.for_microseconds()

    def _time(self):
        return time.time()

    async def get_last_traded_prices(self, trading_pairs: List[str], domain: Optional[str] = None) -> Dict[str, float]:
        return await self._connector.get_last_traded_prices(trading_pairs=trading_pairs)

    async def get_funding_info(self, trading_pair: str) -> FundingInfo:
        funding_info_response = await self._request_complete_funding_info(trading_pair)
        market_info: Dict[str, Any] = funding_info_response["markets"][trading_pair]
        funding_info = FundingInfo(
            trading_pair=trading_pair,
            index_price=Decimal(str(market_info["oraclePrice"])),
            mark_price=Decimal(str(market_info["oraclePrice"])),
            next_funding_utc_timestamp=self._next_funding_time(),
            rate=Decimal(str(market_info["nextFundingRate"])),
        )
        return funding_info

    async def _subscribe_channels(self, ws: WSAssistant):
        try:
            for trading_pair in self._trading_pairs:
                subscribe_orderbook_request: WSJSONRequest = WSJSONRequest(
                    payload={
                        "type": CONSTANTS.WS_TYPE_SUBSCRIBE,
                        "channel": CONSTANTS.WS_CHANNEL_ORDERBOOK,
                        "id": trading_pair,
                    },
                    is_auth_required=False,
                )
                subscribe_trades_request: WSJSONRequest = WSJSONRequest(
                    payload={
                        "type": CONSTANTS.WS_TYPE_SUBSCRIBE,
                        "channel": CONSTANTS.WS_CHANNEL_TRADES,
                        "id": trading_pair,
                    },
                    is_auth_required=False,
                )
                subscribe_markets_request: WSJSONRequest = WSJSONRequest(
                    payload={
                        "type": CONSTANTS.WS_TYPE_SUBSCRIBE,
                        "channel": CONSTANTS.WS_CHANNEL_MARKETS,
                        "id": trading_pair,
                    },
                    is_auth_required=False,
                )
                await ws.send(subscribe_orderbook_request)
                await ws.send(subscribe_trades_request)
                await ws.send(subscribe_markets_request)
            self.logger().info("Subscribed to public orderbook and trade channels...")
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().exception("Unexpected error occurred subscribing to order book trading and delta streams...")
            raise

    def _channel_originating_message(self, event_message: Dict[str, Any]) -> str:
        channel = ""
        if "channel" in event_message:
            event_channel = event_message["channel"]
            event_type = event_message["type"]
            if event_channel == CONSTANTS.WS_CHANNEL_TRADES:
                channel = self._trade_messages_queue_key
            elif event_channel == CONSTANTS.WS_CHANNEL_ORDERBOOK:
                if event_type == CONSTANTS.WS_TYPE_SUBSCRIBED:
                    channel = self._snapshot_messages_queue_key
                if event_type == CONSTANTS.WS_TYPE_CHANNEL_DATA:
                    channel = self._diff_messages_queue_key
            elif event_channel == CONSTANTS.WS_CHANNEL_MARKETS:
                channel = self._funding_info_messages_queue_key
        return channel

    async def _make_order_book_message(
            self,
            raw_message: Dict[str, Any],
            message_queue: asyncio.Queue,
            bids: List[Tuple[float, float]],
            asks: List[Tuple[float, float]],
            message_type: OrderBookMessageType,
    ):
        symbol = raw_message["id"]
        trading_pair = await self._connector.trading_pair_associated_to_exchange_symbol(symbol)

        timestamp_s = self._time()
        update_id = self._nonce_provider.get_tracking_nonce(timestamp=timestamp_s)

        order_book_message_content = {
            "trading_pair": trading_pair,
            "update_id": update_id,
            "bids": bids,
            "asks": asks,
        }
        message = OrderBookMessage(
            message_type=message_type,
            content=order_book_message_content,
            timestamp=timestamp_s,
        )
        message_queue.put_nowait(message)

    async def _parse_order_book_snapshot_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        if raw_message["type"] in ["subscribed", "channel_data"]:
            bids, asks = self._get_bids_and_asks_from_snapshot(raw_message["contents"])
            await self._make_order_book_message(
                raw_message=raw_message,
                message_queue=message_queue,
                bids=bids,
                asks=asks,
                message_type=OrderBookMessageType.SNAPSHOT,
            )

    async def _parse_order_book_diff_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        if raw_message["type"] in ["subscribed", "channel_data"]:
            bids, asks = self._get_bids_and_asks_from_diff(raw_message["contents"])
            await self._make_order_book_message(
                raw_message=raw_message,
                message_queue=message_queue,
                bids=bids,
                asks=asks,
                message_type=OrderBookMessageType.DIFF,
            )

    async def _parse_trade_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        if raw_message["type"] == "channel_data":
            symbol = raw_message["id"]
            trading_pair = await self._connector.trading_pair_associated_to_exchange_symbol(symbol)

            trade_updates = raw_message["contents"]["trades"]

            for trade_data in trade_updates:
                ts_ms = dp.parse(trade_data["createdAt"]).timestamp() * 1e3
                trade_type = float(TradeType.BUY.value) if trade_data["side"] == "BUY" else float(TradeType.SELL.value)
                message_content = {
                    "trade_id": ts_ms,
                    "trading_pair": trading_pair,
                    "trade_type": trade_type,
                    "amount": trade_data["size"],
                    "price": trade_data["price"],
                }
                trade_message = OrderBookMessage(
                    message_type=OrderBookMessageType.TRADE,
                    content=message_content,
                    timestamp=ts_ms * 1e-3,
                )
                message_queue.put_nowait(trade_message)

    async def _parse_funding_info_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        if raw_message["type"] == "channel_data":
            print(raw_message)
            for trading_pair in raw_message["contents"]["markets"].keys():
                if trading_pair in self._trading_pairs:
                    market_info = raw_message["contents"]["markets"][trading_pair]

                    if any(
                            info in ["oraclePrice", "nextFundingRate", "nextFundingAt"]
                            for info in market_info.keys()
                    ):

                        info_update = FundingInfoUpdate(trading_pair)
                        if "oraclePrice" in market_info.keys():
                            info_update.index_price = Decimal(market_info["oraclePrice"])
                            info_update.mark_price = Decimal(market_info["oraclePrice"])
                        if "nextFundingRate" in market_info.keys():
                            info_update.rate = Decimal(market_info["nextFundingRate"])
                            info_update.next_funding_utc_timestamp = self._next_funding_time(),

                        message_queue.put_nowait(info_update)

    async def _request_complete_funding_info(self, trading_pair: str) -> Dict[str, Any]:
        ex_symbol = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)

        params = {
            "limit": 1,
            "ticker": ex_symbol
        }

        rest_assistant = await self._api_factory.get_rest_assistant()
        endpoint = CONSTANTS.PATH_MARKETS
        url = web_utils.public_rest_url(path_url=endpoint)
        data = await rest_assistant.execute_request(
            url=url,
            throttler_limit_id=endpoint,
            params=params,
            method=RESTMethod.GET,
        )
        return data

    async def _order_book_snapshot(self, trading_pair: str) -> OrderBookMessage:
        snapshot_response = await self._request_order_book_snapshot(trading_pair)

        timestamp = self._time()
        update_id = self._nonce_provider.get_tracking_nonce(timestamp=timestamp)

        bids, asks = self._get_bids_and_asks_from_snapshot(snapshot_response)
        order_book_message_content = {
            "trading_pair": trading_pair,
            "update_id": update_id,
            "bids": bids,
            "asks": asks,
        }
        snapshot_msg: OrderBookMessage = OrderBookMessage(
            message_type=OrderBookMessageType.SNAPSHOT,
            content=order_book_message_content,
            timestamp=timestamp,
        )

        return snapshot_msg

    async def _request_order_book_snapshot(self, trading_pair: str) -> Dict[str, Any]:
        rest_assistant = await self._api_factory.get_rest_assistant()
        endpoint = CONSTANTS.PATH_SNAPSHOT
        ex_symbol = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        url = web_utils.public_rest_url(path_url=endpoint + "/" + ex_symbol)
        data = await rest_assistant.execute_request(
            url=url,
            throttler_limit_id=endpoint,
            method=RESTMethod.GET,
        )

        return data

    @staticmethod
    def _get_bids_and_asks_from_snapshot(
            snapshot: Dict[str, List[Dict[str, Union[str, int, float]]]]
    ) -> Tuple[List[Tuple[float, float]], List[Tuple[float, float]]]:

        bids = [(Decimal(bid["price"]), Decimal(bid["size"])) for bid in snapshot["bids"]]
        asks = [(Decimal(ask["price"]), Decimal(ask["size"])) for ask in snapshot["asks"]]

        return bids, asks

    @staticmethod
    def _get_bids_and_asks_from_diff(
            diff: Dict[str, List[Dict[str, Union[str, int, float]]]]
    ) -> Tuple[List[Tuple[float, float]], List[Tuple[float, float]]]:

        bids = [(Decimal(bid[0]), Decimal(bid[1])) for bid in diff.get("bids", [])]
        asks = [(Decimal(ask[0]), Decimal(ask[1])) for ask in diff.get("asks", [])]

        return bids, asks

    async def _connected_websocket_assistant(self) -> WSAssistant:
        ws: WSAssistant = await self._api_factory.get_ws_assistant()
        await ws.connect(ws_url=CONSTANTS.DYDX_V4_WS_URL, ping_timeout=CONSTANTS.HEARTBEAT_INTERVAL)
        return ws

    async def _request_order_book_snapshots(self, output: asyncio.Queue):
        pass  # unused

    def _next_funding_time(self) -> int:
        """
        Funding settlement occurs every 1 hours as mentioned in https://hyperliquid.gitbook.io/hyperliquid-docs/trading/funding
        """
        return ((time.time() // 3600) + 1) * 3600

    @classmethod
    def _get_next_subscribe_id(cls) -> int:
        """Get the next subscription ID and increment the counter."""
        subscribe_id = cls._next_subscribe_id
        cls._next_subscribe_id += 1
        return subscribe_id

    async def subscribe_to_trading_pair(self, trading_pair: str) -> bool:
        """
        Subscribe to order book channels for a single trading pair dynamically.

        :param trading_pair: The trading pair to subscribe to.
        :return: True if subscription was successful, False otherwise.
        """
        if self._ws_assistant is None:
            self.logger().warning(
                f"Cannot subscribe to {trading_pair}: WebSocket connection not established."
            )
            return False

        try:
            subscribe_orderbook_request: WSJSONRequest = WSJSONRequest(
                payload={
                    "type": CONSTANTS.WS_TYPE_SUBSCRIBE,
                    "channel": CONSTANTS.WS_CHANNEL_ORDERBOOK,
                    "id": trading_pair,
                },
                is_auth_required=False,
            )
            subscribe_trades_request: WSJSONRequest = WSJSONRequest(
                payload={
                    "type": CONSTANTS.WS_TYPE_SUBSCRIBE,
                    "channel": CONSTANTS.WS_CHANNEL_TRADES,
                    "id": trading_pair,
                },
                is_auth_required=False,
            )
            subscribe_markets_request: WSJSONRequest = WSJSONRequest(
                payload={
                    "type": CONSTANTS.WS_TYPE_SUBSCRIBE,
                    "channel": CONSTANTS.WS_CHANNEL_MARKETS,
                    "id": trading_pair,
                },
                is_auth_required=False,
            )

            await self._ws_assistant.send(subscribe_orderbook_request)
            await self._ws_assistant.send(subscribe_trades_request)
            await self._ws_assistant.send(subscribe_markets_request)

            self.add_trading_pair(trading_pair)
            self.logger().info(f"Successfully subscribed to {trading_pair}")
            return True

        except asyncio.CancelledError:
            raise
        except Exception as e:
            self.logger().error(f"Error subscribing to {trading_pair}: {e}")
            return False

    async def unsubscribe_from_trading_pair(self, trading_pair: str) -> bool:
        """
        Unsubscribe from order book channels for a single trading pair dynamically.

        :param trading_pair: The trading pair to unsubscribe from.
        :return: True if unsubscription was successful, False otherwise.
        """
        if self._ws_assistant is None:
            self.logger().warning(
                f"Cannot unsubscribe from {trading_pair}: WebSocket connection not established."
            )
            return False

        try:
            unsubscribe_orderbook_request: WSJSONRequest = WSJSONRequest(
                payload={
                    "type": "unsubscribe",
                    "channel": CONSTANTS.WS_CHANNEL_ORDERBOOK,
                    "id": trading_pair,
                },
                is_auth_required=False,
            )
            unsubscribe_trades_request: WSJSONRequest = WSJSONRequest(
                payload={
                    "type": "unsubscribe",
                    "channel": CONSTANTS.WS_CHANNEL_TRADES,
                    "id": trading_pair,
                },
                is_auth_required=False,
            )
            unsubscribe_markets_request: WSJSONRequest = WSJSONRequest(
                payload={
                    "type": "unsubscribe",
                    "channel": CONSTANTS.WS_CHANNEL_MARKETS,
                    "id": trading_pair,
                },
                is_auth_required=False,
            )

            await self._ws_assistant.send(unsubscribe_orderbook_request)
            await self._ws_assistant.send(unsubscribe_trades_request)
            await self._ws_assistant.send(unsubscribe_markets_request)

            self.remove_trading_pair(trading_pair)
            self.logger().info(f"Successfully unsubscribed from {trading_pair}")
            return True

        except asyncio.CancelledError:
            raise
        except Exception as e:
            self.logger().error(f"Error unsubscribing from {trading_pair}: {e}")
            return False



import sys
from decimal import Decimal

from xrpl.asyncio.transaction.main import _LEDGER_OFFSET

from hummingbot.core.api_throttler.data_types import RateLimit
from hummingbot.core.data_type.in_flight_order import OrderState, OrderType

EXCHANGE_NAME = "xrpl"
DOMAIN = "xrpl"  # This just a placeholder since we don't use domain in xrpl connect at the moment

HBOT_SOURCE_TAG_ID = 19089388
HBOT_ORDER_ID_PREFIX = "hbot"
MAX_ORDER_ID_LEN = 40

# Base URL
DEFAULT_JSON_RPC_URL = "https://xrplcluster.com/"
DEFAULT_WSS_URL = "wss://xrplcluster.com/"

# Websocket channels
TRADE_EVENT_TYPE = "trades"
DIFF_EVENT_TYPE = "diffs"
SNAPSHOT_EVENT_TYPE = "order_book_snapshots"

# Drop definitions
ONE_DROP = Decimal("0.000001")

# Ledger Reserve Fee
WALLET_RESERVE = Decimal("1")
LEDGER_OBJECT_RESERVE = Decimal("0.2")

# Order States
ORDER_STATE = {
    "open": OrderState.OPEN,
    "filled": OrderState.FILLED,
    "partial_filled": OrderState.PARTIALLY_FILLED,
    "canceled": OrderState.CANCELED,
    "rejected": OrderState.FAILED,
}

# Order Types
XRPL_ORDER_TYPE = {
    OrderType.LIMIT: 65536,
    OrderType.LIMIT_MAKER: 65536,
    OrderType.MARKET: 262144,
}

XRPL_SELL_FLAG = 524288

# Market Order Max Slippage
MARKET_ORDER_MAX_SLIPPAGE = Decimal("0.01")

# Order Side
SIDE_BUY = 0
SIDE_SELL = 1

# Orderbook settings
ORDER_BOOK_DEPTH = 100
FETCH_ORDER_BOOK_MAX_RETRY = 3
FETCH_ORDER_BOOK_RETRY_INTERVAL = 5

# Ledger offset for getting order status:
LEDGER_OFFSET = _LEDGER_OFFSET * 2

# Timeout for pending order status check
PENDING_ORDER_STATUS_CHECK_TIMEOUT = 120

# Request Timeout
REQUEST_TIMEOUT = 60

# Rate Limits
# NOTE: We don't have rate limits for xrpl at the moment
RAW_REQUESTS = "RAW_REQUESTS"
NO_LIMIT = sys.maxsize
RATE_LIMITS = [
    RateLimit(limit_id=RAW_REQUESTS, limit=NO_LIMIT, time_interval=1),
]

# Place order retry parameters
PLACE_ORDER_MAX_RETRY = 3
PLACE_ORDER_RETRY_INTERVAL = 5

# Transaction fee multiplier
FEE_MULTIPLIER = 5

# Cancel All Timeout
CANCEL_ALL_TIMEOUT = 600

# Cancel retry parameters
CANCEL_MAX_RETRY = 3
CANCEL_RETRY_INTERVAL = 5

# Verify transaction retry parameters
VERIFY_TRANSACTION_MAX_RETRY = 3
VERIFY_TRANSACTION_RETRY_INTERVAL = 5

# Autofill transaction retry parameters
AUTOFILL_TRANSACTION_MAX_RETRY = 3

# Request retry interval
REQUEST_RETRY_INTERVAL = 5

# Request Orderbook Interval
REQUEST_ORDERBOOK_INTERVAL = 10

# Client refresh interval
CLIENT_REFRESH_INTERVAL = 30

# Websocket configuration
WEBSOCKET_MAX_SIZE_BYTES = 2**22  # 4MB
WEBSOCKET_CONNECTION_TIMEOUT = 30

# XRPL maximum digit for issued currency
XRPL_MAX_DIGIT = 16

# Markets list
MARKETS = {
    "XRP-RLUSD": {
        "base": "XRP",
        "quote": "RLUSD",
        "base_issuer": "",
        "quote_issuer": "rMxCKbEDwqr76QuheSUMdEGf4B9xJ8m5De",
    },
    "XRP-IBTC": {
        "base": "XRP",
        "quote": "iBTC",
        "base_issuer": "",
        "quote_issuer": "rGcyRGrZPaJAZbZDi4NqRFLA5GQH63iFpD",
    },
    "XRP-USDC": {
        "base": "XRP",
        "quote": "USDC",
        "base_issuer": "",
        "quote_issuer": "rGm7WCVp9gb4jZHWTEtGUr4dd74z2XuWhE",
    },
    "IBTC-RLUSD": {
        "base": "iBTC",
        "quote": "RLUSD",
        "base_issuer": "rGcyRGrZPaJAZbZDi4NqRFLA5GQH63iFpD",
        "quote_issuer": "rMxCKbEDwqr76QuheSUMdEGf4B9xJ8m5De",
    },
    "XRP-USD": {
        "base": "XRP",
        "quote": "USD",
        "base_issuer": "",
        "quote_issuer": "rhub8VRN55s94qWKDv6jmDy1pUykJzF3wq",
    },
    "XRP-EUR": {
        "base": "XRP",
        "quote": "EUR",
        "base_issuer": "",
        "quote_issuer": "rhub8VRN55s94qWKDv6jmDy1pUykJzF3wq",
    },
    "XRP-GBP": {
        "base": "XRP",
        "quote": "GBP",
        "base_issuer": "",
        "quote_issuer": "r4GN9eEoz9K4BhMQXe4H1eYNtvtkwGdt8g",
    },
    "XRP-BTC": {
        "base": "XRP",
        "quote": "BTC",
        "base_issuer": "",
        "quote_issuer": "rchGBxcD1A1C2tdxF6papQYZ8kjRKMYcL",
    },
    "XRP-ETH": {
        "base": "XRP",
        "quote": "ETH",
        "base_issuer": "",
        "quote_issuer": "rcA8X3TVMST1n3CJeAdGk1RdRCHii7N2h",
    },
    "XRP-LTC": {
        "base": "XRP",
        "quote": "LTC",
        "base_issuer": "",
        "quote_issuer": "rcRzGWq6Ng3jeYhqnmM4zcWcUh69hrQ8V",
    },
    "XRP-CNY": {
        "base": "XRP",
        "quote": "CNY",
        "base_issuer": "",
        "quote_issuer": "rKiCet8SdvWxPXnAgYarFUXMh1zCPz432Y",
    },
    "XRP-BCH": {
        "base": "XRP",
        "quote": "BCH",
        "base_issuer": "",
        "quote_issuer": "rcyS4CeCZVYvTiKcxj6Sx32ibKwcDHLds",
    },
    "XRP-ETC": {
        "base": "XRP",
        "quote": "ETC",
        "base_issuer": "",
        "quote_issuer": "rDAN8tzydyNfnNf2bfUQY6iR96UbpvNsze",
    },
    "XRP-DSH": {
        "base": "XRP",
        "quote": "DSH",
        "base_issuer": "",
        "quote_issuer": "rcXY84C4g14iFp6taFXjjQGVeHqSCh9RX",
    },
    "XRP-XAU": {
        "base": "XRP",
        "quote": "XAU",
        "base_issuer": "",
        "quote_issuer": "rcoef87SYMJ58NAFx7fNM5frVknmvHsvJ",
    },
    "XRP-SGB": {
        "base": "XRP",
        "quote": "SGB",
        "base_issuer": "",
        "quote_issuer": "rctArjqVvTHihekzDeecKo6mkTYTUSBNc",
    },
    "XRP-USDT": {
        "base": "XRP",
        "quote": "USDT",
        "base_issuer": "",
        "quote_issuer": "rcvxE9PS9YBwxtGg1qNeewV6ZB3wGubZq",
    },
    "XRP-WXRP": {
        "base": "XRP",
        "quote": "WXRP",
        "base_issuer": "",
        "quote_issuer": "rEa5QY8tdbjgitLyfKF1E5Qx3VGgvbUhB3",
    },
    "XRP-GALA": {
        "base": "XRP",
        "quote": "GALA",
        "base_issuer": "",
        "quote_issuer": "rf5YPb9y9P3fTjhxNaZqmrwaj5ar8PG1gM",
    },
    "XRP-FLR": {
        "base": "XRP",
        "quote": "FLR",
        "base_issuer": "",
        "quote_issuer": "rcxJwVnftZzXqyH9YheB8TgeiZUhNo1Eu",
    },
    "XRP-XAH": {
        "base": "XRP",
        "quote": "XAH",
        "base_issuer": "",
        "quote_issuer": "rswh1fvyLqHizBS2awu1vs6QcmwTBd9qiv",
    },
    "USD-XRP": {
        "base": "USD",
        "quote": "XRP",
        "base_issuer": "rhub8VRN55s94qWKDv6jmDy1pUykJzF3wq",
        "quote_issuer": "",
    },
    "USD-EUR": {
        "base": "USD",
        "quote": "EUR",
        "base_issuer": "rhub8VRN55s94qWKDv6jmDy1pUykJzF3wq",
        "quote_issuer": "rhub8VRN55s94qWKDv6jmDy1pUykJzF3wq",
    },
    "USD-GBP": {
        "base": "USD",
        "quote": "GBP",
        "base_issuer": "rhub8VRN55s94qWKDv6jmDy1pUykJzF3wq",
        "quote_issuer": "r4GN9eEoz9K4BhMQXe4H1eYNtvtkwGdt8g",
    },
    "USD-BTC": {
        "base": "USD",
        "quote": "BTC",
        "base_issuer": "rhub8VRN55s94qWKDv6jmDy1pUykJzF3wq",
        "quote_issuer": "rchGBxcD1A1C2tdxF6papQYZ8kjRKMYcL",
    },
    "USD-BCH": {
        "base": "USD",
        "quote": "BCH",
        "base_issuer": "rhub8VRN55s94qWKDv6jmDy1pUykJzF3wq",
        "quote_issuer": "rcyS4CeCZVYvTiKcxj6Sx32ibKwcDHLds",
    },
    "USD-LTC": {
        "base": "USD",
        "quote": "LTC",
        "base_issuer": "rhub8VRN55s94qWKDv6jmDy1pUykJzF3wq",
        "quote_issuer": "rcRzGWq6Ng3jeYhqnmM4zcWcUh69hrQ8V",
    },
    "USD.b-XRP": {
        "base": "USD",
        "quote": "XRP",
        "base_issuer": "rvYAfWj5gh67oV6fW32ZzP3Aw4Eubs59B",
        "quote_issuer": "",
    },
    "USD-USDT": {
        "base": "USD",
        "quote": "USDT",
        "base_issuer": "rhub8VRN55s94qWKDv6jmDy1pUykJzF3wq",
        "quote_issuer": "rcvxE9PS9YBwxtGg1qNeewV6ZB3wGubZq",
    },
    "USD-USDC": {
        "base": "USD",
        "quote": "USDC",
        "base_issuer": "rhub8VRN55s94qWKDv6jmDy1pUykJzF3wq",
        "quote_issuer": "rGm7WCVp9gb4jZHWTEtGUr4dd74z2XuWhE",
    },
    "USD-WXRP": {
        "base": "USD",
        "quote": "WXRP",
        "base_issuer": "rhub8VRN55s94qWKDv6jmDy1pUykJzF3wq",
        "quote_issuer": "rEa5QY8tdbjgitLyfKF1E5Qx3VGgvbUhB3",
    },
    "USD-GALA": {
        "base": "USD",
        "quote": "GALA",
        "base_issuer": "rhub8VRN55s94qWKDv6jmDy1pUykJzF3wq",
        "quote_issuer": "rf5YPb9y9P3fTjhxNaZqmrwaj5ar8PG1gM",
    },
    "USD-FLR": {
        "base": "USD",
        "quote": "FLR",
        "base_issuer": "rhub8VRN55s94qWKDv6jmDy1pUykJzF3wq",
        "quote_issuer": "rcxJwVnftZzXqyH9YheB8TgeiZUhNo1Eu",
    },
    "EUR-XRP": {
        "base": "EUR",
        "quote": "XRP",
        "base_issuer": "rhub8VRN55s94qWKDv6jmDy1pUykJzF3wq",
        "quote_issuer": "",
    },
    "EUR-USD": {
        "base": "EUR",
        "quote": "USD",
        "base_issuer": "rhub8VRN55s94qWKDv6jmDy1pUykJzF3wq",
        "quote_issuer": "rhub8VRN55s94qWKDv6jmDy1pUykJzF3wq",
    },
    "EUR-GBP": {
        "base": "EUR",
        "quote": "GBP",
        "base_issuer": "rhub8VRN55s94qWKDv6jmDy1pUykJzF3wq",
        "quote_issuer": "r4GN9eEoz9K4BhMQXe4H1eYNtvtkwGdt8g",
    },
    "EUR-USD.b": {
        "base": "EUR",
        "quote": "USD",
        "base_issuer": "rhub8VRN55s94qWKDv6jmDy1pUykJzF3wq",
        "quote_issuer": "rvYAfWj5gh67oV6fW32ZzP3Aw4Eubs59B",
    },
    "EUR-BTC": {
        "base": "EUR",
        "quote": "BTC",
        "base_issuer": "rhub8VRN55s94qWKDv6jmDy1pUykJzF3wq",
        "quote_issuer": "rchGBxcD1A1C2tdxF6papQYZ8kjRKMYcL",
    },
    "EUR-BCH": {
        "base": "EUR",
        "quote": "BCH",
        "base_issuer": "rhub8VRN55s94qWKDv6jmDy1pUykJzF3wq",
        "quote_issuer": "rcyS4CeCZVYvTiKcxj6Sx32ibKwcDHLds",
    },
    "EUR-LTC": {
        "base": "EUR",
        "quote": "LTC",
        "base_issuer": "rhub8VRN55s94qWKDv6jmDy1pUykJzF3wq",
        "quote_issuer": "rcRzGWq6Ng3jeYhqnmM4zcWcUh69hrQ8V",
    },
    "EUR-USDT": {
        "base": "EUR",
        "quote": "USDT",
        "base_issuer": "rhub8VRN55s94qWKDv6jmDy1pUykJzF3wq",
        "quote_issuer": "rcvxE9PS9YBwxtGg1qNeewV6ZB3wGubZq",
    },
    "EUR-USDC": {
        "base": "EUR",
        "quote": "USDC",
        "base_issuer": "rhub8VRN55s94qWKDv6jmDy1pUykJzF3wq",
        "quote_issuer": "rGm7WCVp9gb4jZHWTEtGUr4dd74z2XuWhE",
    },
    "EUR-WXRP": {
        "base": "EUR",
        "quote": "WXRP",
        "base_issuer": "rhub8VRN55s94qWKDv6jmDy1pUykJzF3wq",
        "quote_issuer": "rEa5QY8tdbjgitLyfKF1E5Qx3VGgvbUhB3",
    },
    "EUR-GALA": {
        "base": "EUR",
        "quote": "GALA",
        "base_issuer": "rhub8VRN55s94qWKDv6jmDy1pUykJzF3wq",
        "quote_issuer": "rf5YPb9y9P3fTjhxNaZqmrwaj5ar8PG1gM",
    },
    "EUR-FLR": {
        "base": "EUR",
        "quote": "FLR",
        "base_issuer": "rhub8VRN55s94qWKDv6jmDy1pUykJzF3wq",
        "quote_issuer": "rcxJwVnftZzXqyH9YheB8TgeiZUhNo1Eu",
    },
    "SGB-XRP": {
        "base": "SGB",
        "quote": "XRP",
        "base_issuer": "rctArjqVvTHihekzDeecKo6mkTYTUSBNc",
        "quote_issuer": "",
    },
    "ELS-XRP": {
        "base": "ELS",
        "quote": "XRP",
        "base_issuer": "rHXuEaRYnnJHbDeuBH5w8yPh5uwNVh5zAg",
        "quote_issuer": "",
    },
    "USDT-XRP": {
        "base": "USDT",
        "quote": "XRP",
        "base_issuer": "rcvxE9PS9YBwxtGg1qNeewV6ZB3wGubZq",
        "quote_issuer": "",
    },
    "USDC-XRP": {
        "base": "USDC",
        "quote": "XRP",
        "base_issuer": "rGm7WCVp9gb4jZHWTEtGUr4dd74z2XuWhE",
        "quote_issuer": "",
    },
    "SOLO-XRP": {
        "base": "SOLO",
        "quote": "XRP",
        "base_issuer": "rsoLo2S1kiGeCcn6hCUXVrCpGMWLrRrLZz",
        "quote_issuer": "",
    },
    "WXRP-XRP": {
        "base": "WXRP",
        "quote": "XRP",
        "base_issuer": "rEa5QY8tdbjgitLyfKF1E5Qx3VGgvbUhB3",
        "quote_issuer": "",
    },
    "GALA-XRP": {
        "base": "GALA",
        "quote": "XRP",
        "base_issuer": "rf5YPb9y9P3fTjhxNaZqmrwaj5ar8PG1gM",
        "quote_issuer": "",
    },
    "FLR-XRP": {
        "base": "FLR",
        "quote": "XRP",
        "base_issuer": "rcxJwVnftZzXqyH9YheB8TgeiZUhNo1Eu",
        "quote_issuer": "",
    },
    "SOLO-USD": {
        "base": "SOLO",
        "quote": "USD",
        "base_issuer": "rsoLo2S1kiGeCcn6hCUXVrCpGMWLrRrLZz",
        "quote_issuer": "rhub8VRN55s94qWKDv6jmDy1pUykJzF3wq",
    },
    "SOLO-USD.b": {
        "base": "SOLO",
        "quote": "USD",
        "base_issuer": "rsoLo2S1kiGeCcn6hCUXVrCpGMWLrRrLZz",
        "quote_issuer": "rvYAfWj5gh67oV6fW32ZzP3Aw4Eubs59B",
    },
    "ICOIN-XRP": {
        "base": "icoin",
        "quote": "XRP",
        "base_issuer": "rJSTh1VLk52tFC3VRXkNWu7Q4nYmfZv7BZ",
        "quote_issuer": "",
    },
    "CORE-XRP": {
        "base": "CORE",
        "quote": "XRP",
        "base_issuer": "rcoreNywaoz2ZCQ8Lg2EbSLnGuRBmun6D",
        "quote_issuer": "",
    },
    "XMEME-XRP": {
        "base": "XMEME",
        "quote": "XRP",
        "base_issuer": "r4UPddYeGeZgDhSGPkooURsQtmGda4oYQW",
        "quote_issuer": "",
    },
    "CSC-XRP": {
        "base": "CSC",
        "quote": "XRP",
        "base_issuer": "rCSCManTZ8ME9EoLrSHHYKW8PPwWMgkwr",
        "quote_issuer": "",
    },
    "FURY-XRP": {
        "base": "FURY",
        "quote": "XRP",
        "base_issuer": "rnoKi9s9b6WYaNGWQy4qVdnKo6Lj2eHE1D",
        "quote_issuer": "",
    },
    "XSPECTAR-XRP": {
        "base": "xSPECTAR",
        "quote": "XRP",
        "base_issuer": "rh5jzTCdMRCVjQ7LT6zucjezC47KATkuvv",
        "quote_issuer": "",
    },
    "RPR-XRP": {
        "base": "RPR",
        "quote": "XRP",
        "base_issuer": "r3qWgpz2ry3BhcRJ8JE6rxM8esrfhuKp4R",
        "quote_issuer": "",
    },
    "XRDOGE-XRP": {
        "base": "XRdoge",
        "quote": "XRP",
        "base_issuer": "rLqUC2eCPohYvJCEBJ77eCCqVL2uEiczjA",
        "quote_issuer": "",
    },
    "EQUILIBRIUM-XRP": {
        "base": "Equilibrium",
        "quote": "XRP",
        "base_issuer": "rpakCr61Q92abPXJnVboKENmpKssWyHpwu",
        "quote_issuer": "",
    },
    "RLUSD-XRP": {
        "base": "RLUSD",
        "quote": "XRP",
        "base_issuer": "rMxCKbEDwqr76QuheSUMdEGf4B9xJ8m5De",
        "quote_issuer": "",
    },
    "RLUSD-USD": {
        "base": "RLUSD",
        "quote": "USD",
        "base_issuer": "rMxCKbEDwqr76QuheSUMdEGf4B9xJ8m5De",
        "quote_issuer": "rhub8VRN55s94qWKDv6jmDy1pUykJzF3wq",
    },
    "USD-RLUSD": {
        "base": "USD",
        "quote": "RLUSD",
        "base_issuer": "rhub8VRN55s94qWKDv6jmDy1pUykJzF3wq",
        "quote_issuer": "rMxCKbEDwqr76QuheSUMdEGf4B9xJ8m5De",
    },
    "RLUSD-EUR": {
        "base": "RLUSD",
        "quote": "EUR",
        "base_issuer": "rMxCKbEDwqr76QuheSUMdEGf4B9xJ8m5De",
        "quote_issuer": "rhub8VRN55s94qWKDv6jmDy1pUykJzF3wq",
    },
    "EUR-RLUSD": {
        "base": "EUR",
        "quote": "RLUSD",
        "base_issuer": "rhub8VRN55s94qWKDv6jmDy1pUykJzF3wq",
        "quote_issuer": "rMxCKbEDwqr76QuheSUMdEGf4B9xJ8m5De",
    },
    "RLUSD-USDT": {
        "base": "RLUSD",
        "quote": "USDT",
        "base_issuer": "rMxCKbEDwqr76QuheSUMdEGf4B9xJ8m5De",
        "quote_issuer": "rcvxE9PS9YBwxtGg1qNeewV6ZB3wGubZq",
    },
    "USDT-RLUSD": {
        "base": "USDT",
        "quote": "RLUSD",
        "base_issuer": "rcvxE9PS9YBwxtGg1qNeewV6ZB3wGubZq",
        "quote_issuer": "rMxCKbEDwqr76QuheSUMdEGf4B9xJ8m5De",
    },
    "RLUSD-USDC": {
        "base": "RLUSD",
        "quote": "USDC",
        "base_issuer": "rMxCKbEDwqr76QuheSUMdEGf4B9xJ8m5De",
        "quote_issuer": "rGm7WCVp9gb4jZHWTEtGUr4dd74z2XuWhE",
    },
    "USDC-RLUSD": {
        "base": "USDC",
        "quote": "RLUSD",
        "base_issuer": "rGm7WCVp9gb4jZHWTEtGUr4dd74z2XuWhE",
        "quote_issuer": "rMxCKbEDwqr76QuheSUMdEGf4B9xJ8m5De",
    },
    "RLUSD-BTC": {
        "base": "RLUSD",
        "quote": "BTC",
        "base_issuer": "rMxCKbEDwqr76QuheSUMdEGf4B9xJ8m5De",
        "quote_issuer": "rchGBxcD1A1C2tdxF6papQYZ8kjRKMYcL",
    },
    "BTC-RLUSD": {
        "base": "BTC",
        "quote": "RLUSD",
        "base_issuer": "rchGBxcD1A1C2tdxF6papQYZ8kjRKMYcL",
        "quote_issuer": "rMxCKbEDwqr76QuheSUMdEGf4B9xJ8m5De",
    },
    "RLUSD-ETH": {
        "base": "RLUSD",
        "quote": "ETH",
        "base_issuer": "rMxCKbEDwqr76QuheSUMdEGf4B9xJ8m5De",
        "quote_issuer": "rcA8X3TVMST1n3CJeAdGk1RdRCHii7N2h",
    },
    "ETH-RLUSD": {
        "base": "ETH",
        "quote": "RLUSD",
        "base_issuer": "rcA8X3TVMST1n3CJeAdGk1RdRCHii7N2h",
        "quote_issuer": "rMxCKbEDwqr76QuheSUMdEGf4B9xJ8m5De",
    },
    "RLUSD-LTC": {
        "base": "RLUSD",
        "quote": "LTC",
        "base_issuer": "rMxCKbEDwqr76QuheSUMdEGf4B9xJ8m5De",
        "quote_issuer": "rcRzGWq6Ng3jeYhqnmM4zcWcUh69hrQ8V",
    },
    "LTC-RLUSD": {
        "base": "LTC",
        "quote": "RLUSD",
        "base_issuer": "rcRzGWq6Ng3jeYhqnmM4zcWcUh69hrQ8V",
        "quote_issuer": "rMxCKbEDwqr76QuheSUMdEGf4B9xJ8m5De",
    },
    "RLUSD-BCH": {
        "base": "RLUSD",
        "quote": "BCH",
        "base_issuer": "rMxCKbEDwqr76QuheSUMdEGf4B9xJ8m5De",
        "quote_issuer": "rcyS4CeCZVYvTiKcxj6Sx32ibKwcDHLds",
    },
    "BCH-RLUSD": {
        "base": "BCH",
        "quote": "RLUSD",
        "base_issuer": "rcyS4CeCZVYvTiKcxj6Sx32ibKwcDHLds",
        "quote_issuer": "rMxCKbEDwqr76QuheSUMdEGf4B9xJ8m5De",
    },
    "RLUSD-GBP": {
        "base": "RLUSD",
        "quote": "GBP",
        "base_issuer": "rMxCKbEDwqr76QuheSUMdEGf4B9xJ8m5De",
        "quote_issuer": "r4GN9eEoz9K4BhMQXe4H1eYNtvtkwGdt8g",
    },
    "GBP-RLUSD": {
        "base": "GBP",
        "quote": "RLUSD",
        "base_issuer": "r4GN9eEoz9K4BhMQXe4H1eYNtvtkwGdt8g",
        "quote_issuer": "rMxCKbEDwqr76QuheSUMdEGf4B9xJ8m5De",
    },
    "RLUSD-WXRP": {
        "base": "RLUSD",
        "quote": "WXRP",
        "base_issuer": "rMxCKbEDwqr76QuheSUMdEGf4B9xJ8m5De",
        "quote_issuer": "rEa5QY8tdbjgitLyfKF1E5Qx3VGgvbUhB3",
    },
    "WXRP-RLUSD": {
        "base": "WXRP",
        "quote": "RLUSD",
        "base_issuer": "rEa5QY8tdbjgitLyfKF1E5Qx3VGgvbUhB3",
        "quote_issuer": "rMxCKbEDwqr76QuheSUMdEGf4B9xJ8m5De",
    },
    "RLUSD-SOLO": {
        "base": "RLUSD",
        "quote": "SOLO",
        "base_issuer": "rMxCKbEDwqr76QuheSUMdEGf4B9xJ8m5De",
        "quote_issuer": "rsoLo2S1kiGeCcn6hCUXVrCpGMWLrRrLZz",
    },
    "SOLO-RLUSD": {
        "base": "SOLO",
        "quote": "RLUSD",
        "base_issuer": "rsoLo2S1kiGeCcn6hCUXVrCpGMWLrRrLZz",
        "quote_issuer": "rMxCKbEDwqr76QuheSUMdEGf4B9xJ8m5De",
    },
    "RLUSD-GALA": {
        "base": "RLUSD",
        "quote": "GALA",
        "base_issuer": "rMxCKbEDwqr76QuheSUMdEGf4B9xJ8m5De",
        "quote_issuer": "rf5YPb9y9P3fTjhxNaZqmrwaj5ar8PG1gM",
    },
    "GALA-RLUSD": {
        "base": "GALA",
        "quote": "RLUSD",
        "base_issuer": "rf5YPb9y9P3fTjhxNaZqmrwaj5ar8PG1gM",
        "quote_issuer": "rMxCKbEDwqr76QuheSUMdEGf4B9xJ8m5De",
    },
    "RLUSD-FLR": {
        "base": "RLUSD",
        "quote": "FLR",
        "base_issuer": "rMxCKbEDwqr76QuheSUMdEGf4B9xJ8m5De",
        "quote_issuer": "rcxJwVnftZzXqyH9YheB8TgeiZUhNo1Eu",
    },
    "FLR-RLUSD": {
        "base": "FLR",
        "quote": "RLUSD",
        "base_issuer": "rcxJwVnftZzXqyH9YheB8TgeiZUhNo1Eu",
        "quote_issuer": "rMxCKbEDwqr76QuheSUMdEGf4B9xJ8m5De",
    },
    "RLUSD-XAU": {
        "base": "RLUSD",
        "quote": "XAU",
        "base_issuer": "rMxCKbEDwqr76QuheSUMdEGf4B9xJ8m5De",
        "quote_issuer": "rcoef87SYMJ58NAFx7fNM5frVknmvHsvJ",
    },
    "XAU-RLUSD": {
        "base": "XAU",
        "quote": "RLUSD",
        "base_issuer": "rcoef87SYMJ58NAFx7fNM5frVknmvHsvJ",
        "quote_issuer": "rMxCKbEDwqr76QuheSUMdEGf4B9xJ8m5De",
    },
    "IBTC-XRP": {
        "base": "iBTC",
        "quote": "XRP",
        "base_issuer": "rGcyRGrZPaJAZbZDi4NqRFLA5GQH63iFpD",
        "quote_issuer": "",
    },
    "IBTC-USDC": {
        "base": "iBTC",
        "quote": "USDC",
        "base_issuer": "rGcyRGrZPaJAZbZDi4NqRFLA5GQH63iFpD",
        "quote_issuer": "rGm7WCVp9gb4jZHWTEtGUr4dd74z2XuWhE",
    },
    "EUROP-RLUSD": {
        "base": "EUROP",
        "quote": "RLUSD",
        "base_issuer": "rMkEuRii9w9uBMQDnWV5AA43gvYZR9JxVK",
        "quote_issuer": "rMxCKbEDwqr76QuheSUMdEGf4B9xJ8m5De",
    },
    "RLUSD-EUROP": {
        "base": "RLUSD",
        "quote": "EUROP",
        "base_issuer": "rMxCKbEDwqr76QuheSUMdEGf4B9xJ8m5De",
        "quote_issuer": "rMkEuRii9w9uBMQDnWV5AA43gvYZR9JxVK",
    },
    "EUROP-XRP": {
        "base": "EUROP",
        "quote": "XRP",
        "base_issuer": "rMkEuRii9w9uBMQDnWV5AA43gvYZR9JxVK",
        "quote_issuer": "",
    },
    "XRP-EUROP": {
        "base": "XRP",
        "quote": "EUROP",
        "base_issuer": "",
        "quote_issuer": "rMkEuRii9w9uBMQDnWV5AA43gvYZR9JxVK",
    },
}



import asyncio
import hashlib
from decimal import Decimal
from typing import Any, Dict, List, Optional, Tuple

import dateutil.parser as dp
from async_timeout import timeout
from bidict import bidict

from hummingbot.connector.constants import s_decimal_NaN
from hummingbot.connector.exchange.dexalot import (
    dexalot_constants as CONSTANTS,
    dexalot_utils,
    dexalot_web_utils as web_utils,
)
from hummingbot.connector.exchange.dexalot.data_sources.dexalot_data_source import DexalotClient
from hummingbot.connector.exchange.dexalot.dexalot_api_order_book_data_source import DexalotAPIOrderBookDataSource
from hummingbot.connector.exchange.dexalot.dexalot_api_user_stream_data_source import DexalotAPIUserStreamDataSource
from hummingbot.connector.exchange.dexalot.dexalot_auth import DexalotAuth
from hummingbot.connector.exchange_py_base import ExchangePyBase
from hummingbot.connector.gateway.gateway_in_flight_order import GatewayInFlightOrder
from hummingbot.connector.trading_rule import TradingRule
from hummingbot.connector.utils import combine_to_hb_trading_pair, get_new_client_order_id
from hummingbot.core.data_type.cancellation_result import CancellationResult
from hummingbot.core.data_type.common import OrderType, TradeType
from hummingbot.core.data_type.in_flight_order import InFlightOrder, OrderState, OrderUpdate, TradeUpdate
from hummingbot.core.data_type.limit_order import LimitOrder
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.data_type.trade_fee import TokenAmount, TradeFeeBase
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.core.utils.estimate_fee import build_trade_fee
from hummingbot.core.web_assistant.connections.data_types import WSJSONRequest
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory


class DexalotExchange(ExchangePyBase):
    UPDATE_ORDER_STATUS_MIN_INTERVAL = 10.0

    web_utils = web_utils

    def __init__(self,
                 dexalot_api_key: str,
                 dexalot_api_secret: str,
                 balance_asset_limit: Optional[Dict[str, Dict[str, Decimal]]] = None,
                 rate_limits_share_pct: Decimal = Decimal("100"),
                 trading_pairs: Optional[List[str]] = None,
                 trading_required: bool = True,
                 domain: str = CONSTANTS.DEFAULT_DOMAIN,
                 ):
        self.api_key = dexalot_api_key
        self.secret_key = dexalot_api_secret
        self._domain = domain
        self._trading_required = trading_required
        self._trading_pairs = trading_pairs
        self._last_trades_poll_dexalot_timestamp = 1.0

        self._orders_queued_to_create: List[GatewayInFlightOrder] = []
        self._orders_queued_to_cancel: List[GatewayInFlightOrder] = []
        self._queued_orders_task = None

        self._evm_params = {}
        self._tx_client: DexalotClient = self._create_tx_client()

        super().__init__(balance_asset_limit, rate_limits_share_pct)

    @staticmethod
    def dexalot_order_type(order_type: OrderType) -> str:
        return order_type.name.upper()

    @staticmethod
    def to_hb_order_type(dexalot_type: str) -> OrderType:
        return OrderType[dexalot_type]

    @property
    def authenticator(self):
        return DexalotAuth(
            api_key=self.api_key,
            secret_key=self.secret_key,
            time_provider=self._time_synchronizer)

    @property
    def name(self) -> str:
        return self._domain

    @property
    def rate_limits_rules(self):
        return CONSTANTS.RATE_LIMITS

    @property
    def domain(self):
        return self._domain

    @property
    def client_order_id_max_length(self):
        return CONSTANTS.MAX_ORDER_ID_LEN

    @property
    def client_order_id_prefix(self):
        return CONSTANTS.HBOT_ORDER_ID_PREFIX

    @property
    def trading_rules_request_path(self):
        return CONSTANTS.EXCHANGE_INFO_PATH_URL

    @property
    def trading_pairs_request_path(self):
        return CONSTANTS.EXCHANGE_INFO_PATH_URL

    @property
    def check_network_request_path(self):
        return CONSTANTS.PING_PATH_URL

    @property
    def trading_pairs(self):
        return self._trading_pairs

    @property
    def is_cancel_request_in_exchange_synchronous(self) -> bool:
        return False

    @property
    def is_trading_required(self) -> bool:
        return self._trading_required

    async def start_network(self):
        await super().start_network()
        await self._update_trading_rules()
        if self.is_trading_required:
            self._queued_orders_task = safe_ensure_future(self._process_queued_orders())

    async def stop_network(self):
        """
        This function is executed when the connector is stopped. It performs a general cleanup and stops all background
        tasks that require the connection with the exchange to work.
        """
        await super().stop_network()
        if self._queued_orders_task is not None:
            self._queued_orders_task.cancel()
            self._queued_orders_task = None

    def supported_order_types(self):
        return [OrderType.LIMIT, OrderType.LIMIT_MAKER, OrderType.MARKET]

    async def get_all_pairs_prices(self) -> List[Dict[str, str]]:
        # pairs_prices = await self._api_get(path_url=CONSTANTS.ALL_TICKERS_PATH_URL)
        api_factory = self._web_assistants_factory
        ws = await api_factory.get_ws_assistant()
        async with api_factory.throttler.execute_task(limit_id=CONSTANTS.WSS_URL):
            await ws.connect(ws_url=web_utils.wss_url(self._domain), ping_timeout=CONSTANTS.WS_HEARTBEAT_TIME_INTERVAL)
            payload = {
                "type": "marketsnapshotsubscribe",
            }
            subscribe_orderbook_request: WSJSONRequest = WSJSONRequest(payload=payload)
            await ws.send(subscribe_orderbook_request)
        async for msg in ws.iter_messages():
            data = msg.data
            if data is not None and data["type"] == "marketSnapShot":
                price_list = data["data"]
                payload = {
                    "type": "marketsnapshotunsubscribe",
                }
                subscribe_orderbook_request: WSJSONRequest = WSJSONRequest(payload=payload)
                await ws.send(subscribe_orderbook_request)
                await ws.disconnect()
                return price_list

    def _format_evmamount_to_amount(self, trading_pair, base_evm_amount: Decimal, quote_evm_amount: Decimal) -> Tuple:

        base_evmdecimals = self._evm_params[trading_pair].get("base_evmdecimals")
        quote_evmdecimals = self._evm_params[trading_pair].get("quote_evmdecimals")
        base_amount = base_evm_amount * Decimal(f"1e-{base_evmdecimals}")
        quote_amount = quote_evm_amount * Decimal(f"1e-{quote_evmdecimals}")

        return base_amount, quote_amount

    def _format_amount_to_evmamount(self, trading_pair, base_amount: Decimal, quote_amount: Decimal) -> Tuple:

        base_evmdecimals = self._evm_params[trading_pair].get("base_evmdecimals")
        quote_evmdecimals = self._evm_params[trading_pair].get("quote_evmdecimals")
        base_evm_amount = base_amount * Decimal(f"1e{base_evmdecimals}")
        quote_evm_amount = quote_amount * Decimal(f"1e{quote_evmdecimals}")

        return base_evm_amount, quote_evm_amount

    def _is_request_exception_related_to_time_synchronizer(self, request_exception: Exception):
        return False

    def _is_order_not_found_during_status_update_error(self, status_update_exception: Exception) -> bool:
        return False

    def _is_order_not_found_during_cancelation_error(self, cancelation_exception: Exception) -> bool:
        return False

    def _create_web_assistants_factory(self) -> WebAssistantsFactory:
        return web_utils.build_api_factory(
            throttler=self._throttler,
            time_synchronizer=self._time_synchronizer,
            domain=self._domain,
            auth=self._auth)

    def _create_tx_client(self) -> DexalotClient:
        return DexalotClient(
            self.secret_key,
            connector=self,
            domain=self._domain,
            trading_required=self.is_trading_required
        )

    def _create_order_book_data_source(self) -> OrderBookTrackerDataSource:
        return DexalotAPIOrderBookDataSource(
            trading_pairs=self._trading_pairs,
            connector=self,
            domain=self.domain,
            api_factory=self._web_assistants_factory)

    def _create_user_stream_data_source(self) -> UserStreamTrackerDataSource:
        return DexalotAPIUserStreamDataSource(
            auth=self._auth,
            api_factory=self._web_assistants_factory,
            domain=self.domain,
        )

    def _get_fee(self,
                 base_currency: str,
                 quote_currency: str,
                 order_type: OrderType,
                 order_side: TradeType,
                 amount: Decimal,
                 price: Decimal = s_decimal_NaN,
                 is_maker: Optional[bool] = None) -> TradeFeeBase:
        is_maker = order_type is OrderType.LIMIT_MAKER
        trade_base_fee = build_trade_fee(
            exchange=self.name,
            is_maker=is_maker,
            order_side=order_side,
            order_type=order_type,
            amount=amount,
            price=price,
            base_currency=base_currency.upper(),
            quote_currency=quote_currency.upper()
        )
        return trade_base_fee

    def _on_order_creation_failure(
            self,
            order_id: str,
            trading_pair: str,
            amount: Decimal,
            trade_type: TradeType,
            order_type: OrderType,
            price: Optional[Decimal],
            exception: Exception,
    ):
        self.logger().network(
            f"Error submitting {trade_type.name.lower()} {order_type.name.upper()} order to {self.name_cap} for "
            f"{amount} {trading_pair} {price}.",
            exc_info=exception,
            app_warning_msg=f"Failed to submit {trade_type.name.upper()} order to {self.name_cap}. Check API key and network connection."
        )
        self._update_order_after_creation_failure(order_id=order_id, trading_pair=trading_pair)

    def _update_order_after_creation_failure(self, order_id: str, trading_pair: str):
        order_update: OrderUpdate = OrderUpdate(
            client_order_id=order_id,
            trading_pair=trading_pair,
            update_timestamp=self.current_timestamp,
            new_state=OrderState.FAILED,
        )
        self._order_tracker.process_order_update(order_update)
        return order_update

    def batch_order_cancel(self, orders_to_cancel: List[LimitOrder]):
        """
        Issues a batch order cancelation as a single API request for exchanges that implement this feature. The default
        implementation of this method is to send the requests discretely (one by one).
        :param orders_to_cancel: A list of the orders to cancel.
        """
        safe_ensure_future(coro=self._execute_batch_cancel(orders_to_cancel=orders_to_cancel))

    async def cancel_all(self, timeout_seconds: float) -> List[CancellationResult]:
        """
        Cancels all currently active orders. The cancellations are performed in parallel tasks.

        :param timeout_seconds: the maximum time (in seconds) the cancel logic should run

        :return: a list of CancellationResult instances, one for each of the orders to be cancelled
        """
        incomplete_orders = {}
        limit_orders = []
        successful_cancellations = []

        for order in self.in_flight_orders.values():
            if not order.is_done:
                incomplete_orders[order.client_order_id] = order
                limit_orders.append(order.to_limit_order())

        if len(limit_orders) > 0:
            try:
                async with timeout(timeout_seconds):
                    cancellation_results = await self._execute_batch_cancel(orders_to_cancel=limit_orders)
                    for cr in cancellation_results:
                        if cr.success:
                            del incomplete_orders[cr.order_id]
                            successful_cancellations.append(CancellationResult(cr.order_id, True))
            except Exception:
                self.logger().network(
                    "Unexpected error cancelling orders.",
                    exc_info=True,
                    app_warning_msg="Failed to cancel order. Check API key and network connection."
                )
        # Give some time for cancellation events to trigger
        await asyncio.sleep(2)
        failed_cancellations = [CancellationResult(oid, False) for oid in incomplete_orders.keys()]
        return successful_cancellations + failed_cancellations

    async def _execute_batch_cancel(self, orders_to_cancel: List[LimitOrder]) -> List[CancellationResult]:
        results = []
        tracked_orders_to_cancel = []

        for order in orders_to_cancel:
            tracked_order = self._order_tracker.all_updatable_orders.get(order.client_order_id)
            if tracked_order is not None and tracked_order.exchange_order_id:
                tracked_orders_to_cancel.append(tracked_order)
            else:
                results.append(CancellationResult(order_id=order.client_order_id, success=False))

        if len(tracked_orders_to_cancel) > 0:
            results.extend(await self._execute_batch_order_cancel(orders_to_cancel=tracked_orders_to_cancel))

        return results

    async def _execute_batch_order_cancel(self,
                                          orders_to_cancel: List[GatewayInFlightOrder]) -> List[CancellationResult]:
        try:
            async with self._throttler.execute_task(limit_id=CONSTANTS.UID_REQUEST_WEIGHT):
                cancelation_results = []

                cancel_transaction_hash = await self._tx_client.cancel_order_list(orders_to_cancel=orders_to_cancel)
                for cancel_order_result in orders_to_cancel:
                    success = True
                    order_update: OrderUpdate = OrderUpdate(
                        client_order_id=cancel_order_result.client_order_id,
                        trading_pair=cancel_order_result.trading_pair,
                        update_timestamp=self.current_timestamp,
                        new_state=(OrderState.CANCELED
                                   if self.is_cancel_request_in_exchange_synchronous
                                   else OrderState.PENDING_CANCEL),
                        misc_updates={"cancelation_transaction_hash": cancel_transaction_hash},
                    )
                    self._order_tracker.process_order_update(order_update)
                    cancelation_results.append(
                        CancellationResult(order_id=cancel_order_result.client_order_id, success=success)
                    )
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().error(
                f"Failed to cancel orders {', '.join([o.client_order_id for o in orders_to_cancel])}",
                exc_info=True,
            )
            cancelation_results = [
                CancellationResult(order_id=order.client_order_id, success=False)
                for order in orders_to_cancel
            ]

        return cancelation_results

    async def _place_cancel(self, order_id: str, tracked_order: GatewayInFlightOrder):
        # Not required because of _execute_order_cancel redefinition
        raise NotImplementedError

    async def _execute_order_cancel(self, order: GatewayInFlightOrder) -> str:
        # Order cancelation requests for single orders are queued to be executed in batch if possible
        self._orders_queued_to_cancel.append(order)
        return None

    async def _place_order(self, order_id: str, trading_pair: str, amount: Decimal, trade_type: TradeType,
                           order_type: OrderType, price: Decimal, **kwargs) -> Tuple[str, float]:
        # Not required because of _place_order_and_process_update redefinition
        raise NotImplementedError

    async def _place_order_and_process_update(self, order: GatewayInFlightOrder, **kwargs) -> str:
        # Order creation requests for single orders are queued to be executed in batch if possible
        self._orders_queued_to_create.append(order)
        return None

    def buy(self,
            trading_pair: str,
            amount: Decimal,
            order_type=OrderType.LIMIT,
            price: Decimal = s_decimal_NaN,
            **kwargs) -> str:
        """
        Creates a promise to create a buy order using the parameters

        :param trading_pair: the token pair to operate with
        :param amount: the order amount
        :param order_type: the type of order to create (MARKET, LIMIT, LIMIT_MAKER)
        :param price: the order price

        :return: the id assigned by the connector to the order (the client id)
        """
        order_id = get_new_client_order_id(
            is_buy=True,
            trading_pair=trading_pair,
            hbot_order_id_prefix=self.client_order_id_prefix,
            max_id_len=self.client_order_id_max_length
        )
        md5 = hashlib.sha256()
        md5.update(order_id.encode('utf-8'))
        hex_order_id = f"0x{md5.hexdigest()}"

        if order_type is OrderType.MARKET:
            price = Decimal(0)

        safe_ensure_future(self._create_order(
            trade_type=TradeType.BUY,
            order_id=hex_order_id,
            trading_pair=trading_pair,
            amount=amount,
            order_type=order_type,
            price=price,
            **kwargs))
        return hex_order_id

    def sell(self,
             trading_pair: str,
             amount: Decimal,
             order_type: OrderType = OrderType.LIMIT,
             price: Decimal = s_decimal_NaN,
             **kwargs) -> str:
        """
        Creates a promise to create a sell order using the parameters.
        :param trading_pair: the token pair to operate with
        :param amount: the order amount
        :param order_type: the type of order to create (MARKET, LIMIT, LIMIT_MAKER)
        :param price: the order price
        :return: the id assigned by the connector to the order (the client id)
        """
        order_id = get_new_client_order_id(
            is_buy=False,
            trading_pair=trading_pair,
            hbot_order_id_prefix=self.client_order_id_prefix,
            max_id_len=self.client_order_id_max_length
        )
        md5 = hashlib.sha256()
        md5.update(order_id.encode('utf-8'))
        hex_order_id = f"0x{md5.hexdigest()}"
        if order_type is OrderType.MARKET:
            price = Decimal(0)

        safe_ensure_future(self._create_order(
            trade_type=TradeType.SELL,
            order_id=hex_order_id,
            trading_pair=trading_pair,
            amount=amount,
            order_type=order_type,
            price=price,
            **kwargs))
        return hex_order_id

    async def _execute_batch_inflight_order_cancel_and_create(
            self,
            orders_to_cancel: List[LimitOrder],
            inflight_orders_to_create: List[GatewayInFlightOrder]
    ):
        tracked_orders_to_cancel = []
        for order in orders_to_cancel:
            tracked_order = self._order_tracker.all_updatable_orders.get(order.client_order_id)
            if tracked_order is not None and tracked_order.exchange_order_id:
                tracked_orders_to_cancel.append(tracked_order)
        try:
            async with self._throttler.execute_task(limit_id=CONSTANTS.UID_REQUEST_WEIGHT):

                transaction_hash = await self._tx_client.cancel_and_add_order_list(
                    orders_to_cancel = tracked_orders_to_cancel,
                    order_list=inflight_orders_to_create
                )
                for cancel_order_result in tracked_orders_to_cancel:
                    order_update: OrderUpdate = OrderUpdate(
                        client_order_id=cancel_order_result.client_order_id,
                        trading_pair=cancel_order_result.trading_pair,
                        update_timestamp=self.current_timestamp,
                        new_state=(OrderState.CANCELED
                                   if self.is_cancel_request_in_exchange_synchronous
                                   else OrderState.PENDING_CANCEL),
                        misc_updates={"cancelation_transaction_hash": transaction_hash}
                    )
                    self._order_tracker.process_order_update(order_update)
                for in_flight_order in inflight_orders_to_create:
                    order_update: OrderUpdate = OrderUpdate(
                        client_order_id=in_flight_order.client_order_id,
                        exchange_order_id=None,
                        trading_pair=in_flight_order.trading_pair,
                        update_timestamp=self.current_timestamp,
                        new_state=in_flight_order.current_state,
                        misc_updates={"creation_transaction_hash": transaction_hash},
                    )
                    self.logger().debug(
                        f"\nCreated order {in_flight_order.client_order_id}  with TX {transaction_hash}")
                    self._order_tracker.process_order_update(order_update)

        except asyncio.CancelledError:
            raise
        except Exception as ex:
            self.logger().error(
                f"Failed to cancel orders {', '.join([o.client_order_id for o in orders_to_cancel])}",
                exc_info=True,
            )
            self.logger().network("Batch order create failed.")
            for order in inflight_orders_to_create:
                self._on_order_creation_failure(
                    order_id=order.client_order_id,
                    trading_pair=order.trading_pair,
                    amount=order.amount,
                    trade_type=order.trade_type,
                    order_type=order.order_type,
                    price=order.price,
                    exception=ex,
                )

    async def _format_trading_rules(self, exchange_info_dict: List) -> List[TradingRule]:
        trading_pair_rules = exchange_info_dict
        retval = []
        for rule in filter(dexalot_utils.is_exchange_information_valid, trading_pair_rules):
            try:
                trading_pair = await self.trading_pair_associated_to_exchange_symbol(symbol=rule.get("pair"))
                min_order_size = Decimal(f"1e-{rule['basedisplaydecimals']}")
                min_price_inc = Decimal(f"1e-{rule['quotedisplaydecimals']}")
                min_notional = Decimal(rule['mintrade_amnt'])
                retval.append(
                    TradingRule(trading_pair,
                                min_order_size=min_order_size,
                                min_price_increment=min_price_inc,
                                min_base_amount_increment=min_order_size,
                                min_notional_size=min_notional))

                self._evm_params[trading_pair] = {
                    "base_coin": rule["base"],
                    "quote_coin": rule["quote"],
                    "base_evmdecimals": Decimal(rule["quote_evmdecimals"]),
                    "quote_evmdecimals": Decimal(rule["base_evmdecimals"]),
                }

            except Exception:
                self.logger().exception(f"Error parsing the trading pair rule {rule}. Skipping.")
        return retval

    async def _update_trading_fees(self):
        """
        Update fees information from the exchange
        """
        pass

    async def _user_stream_event_listener(self):
        """
        Listens to messages from _user_stream_tracker.user_stream queue.
        Traders, Orders, and Balance updates from the WS.
        """
        async for event_message in self._iter_user_event_queue():
            try:
                channel: str = event_message.get("type", None)
                if channel == CONSTANTS.USER_TRADES_ENDPOINT_NAME:
                    safe_ensure_future(self._process_trade_message(event_message))
                elif channel == CONSTANTS.USER_ORDERS_ENDPOINT_NAME:
                    self._process_order_message(event_message)

            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().error(
                    "Unexpected error in user stream listener loop.", exc_info=True)
                await self._sleep(5.0)

    def _create_trade_update_with_order_fill_data(
            self,
            order_fill: Dict[str, Any],
            order: InFlightOrder):

        is_maker = True if order_fill.get("addressMaker", "") == self.api_key else False
        takerSide = order_fill.get("takerSide")
        if is_maker:
            fee_amount = Decimal(order_fill.get("feeMaker", 0))
            if takerSide == "BUY":
                fee_asset = order.quote_asset
            else:
                fee_asset = order.base_asset
        else:
            fee_amount = Decimal(order_fill.get("feeTaker", 0))
            if takerSide == "BUY":
                fee_asset = order.base_asset
            else:
                fee_asset = order.quote_asset

        fee = TradeFeeBase.new_spot_fee(
            fee_schema=self.trade_fee_schema(),
            trade_type=order.trade_type,
            percent_token=fee_asset.upper(),
            flat_fees=[TokenAmount(
                amount=Decimal(fee_amount),
                token=fee_asset.upper()
            )]
        )

        trade_update = TradeUpdate(
            trade_id=str(order_fill["execId"]),
            client_order_id=order.client_order_id,
            exchange_order_id=order.exchange_order_id,
            trading_pair=order.trading_pair,
            fee=fee,
            fill_base_amount=Decimal(order_fill["quantity"]),
            fill_quote_amount=Decimal(order_fill["quantity"]) * Decimal(order_fill["price"]),
            fill_price=Decimal(order_fill["price"]),
            fill_timestamp=order_fill["blockTimestamp"],
        )
        return trade_update

    async def _process_trade_message(self, trade: Dict[str, Any], client_order_id: Optional[str] = None):

        exchange_order_id = trade["data"].get("makerOrder", "") \
            if trade["data"].get("addressMaker", "") == self.api_key else trade["data"].get("takerOrder", "")
        all_orders = self._order_tracker.all_fillable_orders
        self._calculate_available_balance_from_trades(trade["data"])
        try:
            for k, v in all_orders.items():
                await v.get_exchange_order_id()
        except Exception:
            pass
        _cli_tracked_orders = [o for o in all_orders.values() if exchange_order_id == o.exchange_order_id]
        if len(_cli_tracked_orders) == 0 or _cli_tracked_orders[0] is None:
            order_update: OrderUpdate = await self._request_order_status(tracked_order=None,
                                                                         exchange_order_id=exchange_order_id)
            # NOTE: Untracked order
            if order_update is None:
                self.logger().debug(f"Received untracked order with exchange order id of {exchange_order_id}")
                return
            client_order_id = order_update.client_order_id
            tracked_order = self._order_tracker.all_fillable_orders.get(client_order_id)
        else:
            tracked_order = _cli_tracked_orders[0]

        if tracked_order is None:
            self.logger().debug(f"Ignoring trade message with id {client_order_id}: not in in_flight_orders.")
        else:
            trade_update = self._create_trade_update_with_order_fill_data(
                order_fill=trade["data"],
                order=tracked_order)
            self._order_tracker.process_trade_update(trade_update)

    def _create_order_update_with_order_status_data(self, order_status: Dict[str, Any], order: InFlightOrder):
        client_order_id = str(order_status.get("clientOrderId", ""))
        order.update_exchange_order_id(order_status["orderId"])
        order_update = OrderUpdate(
            trading_pair=order.trading_pair,
            update_timestamp=int(order_status["blockTimestamp"]),
            new_state=CONSTANTS.ORDER_STATE[order_status["status"]],
            client_order_id=client_order_id,
            exchange_order_id=str(order_status["orderId"]),
        )
        return order_update

    def _process_order_message(self, raw_msg: Dict[str, Any]):
        order_msg = raw_msg.get("data", {})
        client_order_id = str(order_msg.get("clientOrderId", ""))
        tracked_order = self._order_tracker.all_updatable_orders.get(client_order_id)
        self._calculate_available_balance_from_orders(order_msg)
        if not tracked_order:
            self.logger().debug(f"Ignoring order message with id {client_order_id}: not in in_flight_orders.")
            return

        order_update = self._create_order_update_with_order_status_data(order_status=order_msg, order=tracked_order)
        self._order_tracker.process_order_update(order_update=order_update)

    def _calculate_available_balance_from_trades(self, trade_msg: Dict):
        base_coin = trade_msg["pair"].split("/")[0].upper()
        quote_coin = trade_msg["pair"].split("/")[1].upper()

        is_maker = True if trade_msg.get("addressMaker", "") == self.api_key else False
        takerSide = trade_msg.get("takerSide")
        if is_maker:
            if takerSide == "BUY":
                base_collateral_value = Decimal(trade_msg["quantity"])
                quote_collateral_value = Decimal(trade_msg.get("quantity")) * Decimal(trade_msg.get("price"))
                self._account_available_balances[quote_coin] += quote_collateral_value

                self._account_balances[quote_coin] += quote_collateral_value
                self._account_balances[base_coin] -= base_collateral_value

            else:
                base_collateral_value = Decimal(trade_msg["quantity"])
                quote_collateral_value = Decimal(trade_msg.get("quantity")) * Decimal(trade_msg.get("price"))
                self._account_available_balances[base_coin] += base_collateral_value

                self._account_balances[quote_coin] -= quote_collateral_value
                self._account_balances[base_coin] += base_collateral_value
        else:
            if takerSide == "BUY":
                base_collateral_value = Decimal(trade_msg["quantity"])
                quote_collateral_value = Decimal(trade_msg.get("quantity")) * Decimal(trade_msg.get("price"))
                self._account_available_balances[base_coin] += base_collateral_value

                self._account_balances[quote_coin] -= quote_collateral_value
                self._account_balances[base_coin] += base_collateral_value

            else:
                base_collateral_value = Decimal(trade_msg["quantity"])
                quote_collateral_value = Decimal(trade_msg.get("quantity")) * Decimal(trade_msg.get("price"))
                self._account_available_balances[quote_coin] += quote_collateral_value

                self._account_balances[quote_coin] += quote_collateral_value
                self._account_balances[base_coin] -= base_collateral_value

    def _calculate_available_balance_from_orders(self, order_msg: Dict):
        if order_msg.get("pair"):
            base_coin = order_msg["pair"].split("/")[0].upper()
            quote_coin = order_msg["pair"].split("/")[1].upper()
            if order_msg["status"] in ["NEW", 0]:
                if order_msg["side"] == "BUY" or order_msg["side"] == 0:
                    quote_collateral_value = Decimal(order_msg.get("price")) * Decimal(order_msg.get("quantity"))
                    self._account_available_balances[quote_coin] -= quote_collateral_value
                else:
                    base_collateral_value = Decimal(order_msg["quantity"])
                    self._account_available_balances[base_coin] -= base_collateral_value
            # Partial status used to update _account_available_balances during update_balance
            if order_msg["status"] in [2]:
                if order_msg["side"] == 0:  # BUY
                    quote_collateral_unfilled_value = \
                        Decimal(order_msg["price"]) * Decimal(order_msg["quantity"]) - Decimal(order_msg["totalamount"])
                    self._account_available_balances[quote_coin] -= quote_collateral_unfilled_value
                else:
                    base_collateral_unfilled_value = Decimal(order_msg["quantity"]) - Decimal(
                        order_msg["quantityfilled"])
                    self._account_available_balances[base_coin] -= base_collateral_unfilled_value
            if order_msg["status"] in ["CANCELED", 4]:
                if order_msg["side"] == "BUY" or order_msg["side"] == 0:
                    quote_collateral_value = Decimal(order_msg.get("price")) * Decimal(order_msg.get("quantity"))
                    quote_filled_value = Decimal(order_msg.get("totalamount") or order_msg.get("totalAmount"))
                    self._account_available_balances[quote_coin] += quote_collateral_value
                    self._account_available_balances[quote_coin] -= quote_filled_value
                else:
                    base_collateral_value = Decimal(order_msg["quantity"])
                    base_filled_value = Decimal(order_msg["quantityfilled"])
                    self._account_available_balances[base_coin] += base_collateral_value
                    self._account_available_balances[base_coin] -= base_filled_value

    async def _all_trade_updates_for_order(self, order: InFlightOrder) -> List[TradeUpdate]:
        trade_updates = []

        if order.exchange_order_id is not None:
            exchange_order_id = order.exchange_order_id
            trading_pair = await self.exchange_symbol_associated_to_pair(trading_pair=order.trading_pair)
            all_fills_response = await self._api_get(
                path_url=CONSTANTS.MY_TRADES_PATH_URL,
                params={
                    "orderid": exchange_order_id
                },
                is_auth_required=True,
                limit_id=CONSTANTS.IP_REQUEST_WEIGHT)

            for trade in all_fills_response:
                exchange_order_id = str(trade["orderid"])
                fee = TradeFeeBase.new_spot_fee(
                    fee_schema=self.trade_fee_schema(),
                    trade_type=order.trade_type,
                    percent_token=trade["feeunit"].upper(),
                    flat_fees=[TokenAmount(amount=Decimal(trade["fee"]), token=trade["feeunit"].upper())]
                )
                trade_update = TradeUpdate(
                    trade_id=str(trade["execid"]),
                    client_order_id=order.client_order_id,
                    exchange_order_id=exchange_order_id,
                    trading_pair=trading_pair,
                    fee=fee,
                    fill_base_amount=Decimal(trade["quantity"]),
                    fill_quote_amount=Decimal(trade["quantity"]) * Decimal(trade["price"]),
                    fill_price=Decimal(trade["price"]),
                    fill_timestamp=dp.parse(trade["ts"]).timestamp(),
                )
                trade_updates.append(trade_update)

        return trade_updates

    async def _request_order_status(self, tracked_order: InFlightOrder, exchange_order_id=None) -> OrderUpdate:
        try:
            if not exchange_order_id:
                exchange_order_id = await tracked_order.get_exchange_order_id()
        except asyncio.TimeoutError:
            self.logger().warning(
                f"Error fetching status update for the lost order {tracked_order.client_order_id}: TimeoutError.")
            order_update = self._update_order_after_creation_failure(
                tracked_order.client_order_id,
                tracked_order.trading_pair
            )
            return order_update
        if not tracked_order:
            all_fillable_orders_by_exchange_order_id = {
                order.exchange_order_id: order for order in self._order_tracker.all_fillable_orders.values()
            }
            tracked_order = all_fillable_orders_by_exchange_order_id.get(exchange_order_id)

        updated_order_data = await self._api_get(
            path_url=CONSTANTS.ORDER_PATH_URL.format(exchange_order_id),
            params={},
            is_auth_required=True,
            limit_id=CONSTANTS.IP_REQUEST_WEIGHT)
        client_order_id = updated_order_data.get("clientOrderId")
        tracked_order = self._order_tracker.all_fillable_orders.get(
            client_order_id) if not tracked_order else tracked_order
        if not tracked_order:
            self.logger().debug(f"Ignoring order message with id {client_order_id}: not in in_flight_orders.")
            return

        new_state = CONSTANTS.ORDER_STATE[updated_order_data["status"]]

        order_update = OrderUpdate(
            client_order_id=client_order_id or tracked_order.client_order_id,
            exchange_order_id=str(exchange_order_id),
            trading_pair=tracked_order.trading_pair,
            update_timestamp=dp.parse(updated_order_data["timestamp"]).timestamp(),
            new_state=new_state,
        )

        return order_update

    async def _update_balances(self):
        self._account_balances, self._account_available_balances = await self._tx_client.get_balances(
            self._account_balances, self._account_available_balances
        )

        open_orders = await self._api_get(
            path_url=CONSTANTS.ORDERS_PATH_URL,
            is_auth_required=True,
            limit_id=CONSTANTS.IP_REQUEST_WEIGHT
        )
        for order_msg in open_orders["rows"]:
            self._calculate_available_balance_from_orders(order_msg)

    def _initialize_trading_pair_symbols_from_exchange_info(self, exchange_info: List):
        mapping = bidict()
        for symbol_data in filter(dexalot_utils.is_exchange_information_valid, exchange_info):
            mapping[symbol_data["pair"]] = combine_to_hb_trading_pair(base=symbol_data["base"].upper(),
                                                                      quote=symbol_data["quote"].upper())
        self._set_trading_pair_symbol_map(mapping)

    async def _process_queued_orders(self):
        while True:
            try:
                # Executing the batch cancelation and creation process shielded from this async task to isolate the
                # creation/cancelation process from network disconnections (network disconnections cancel this task)
                task = asyncio.create_task(self._cancel_and_create_queued_orders())
                await asyncio.shield(task)
                sleep_time = (self.clock.tick_size * 0.5
                              if self.clock is not None
                              else self._orders_processing_delta_time)
                await self._sleep(sleep_time)
            except NotImplementedError:
                raise
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().exception("Unexpected error while processing queued individual orders", exc_info=True)
                await self._sleep(self.clock.tick_size * 0.5)

    async def _cancel_and_create_queued_orders(self):
        if len(self._orders_queued_to_cancel) > 0 or len(self._orders_queued_to_create) > 0:
            cancel_orders = [order.to_limit_order() for order in self._orders_queued_to_cancel]
            add_orders = self._orders_queued_to_create
            self._orders_queued_to_cancel = []
            self._orders_queued_to_create = []
            await self._execute_batch_inflight_order_cancel_and_create(
                orders_to_cancel=cancel_orders,
                inflight_orders_to_create=add_orders
            )

    async def _get_last_traded_price(self, trading_pair: str) -> float:
        ex_symbol = await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        api_factory = self._web_assistants_factory
        ws = await api_factory.get_ws_assistant()
        async with api_factory.throttler.execute_task(limit_id=CONSTANTS.WSS_URL):
            await ws.connect(ws_url=web_utils.wss_url(self._domain), ping_timeout=CONSTANTS.WS_HEARTBEAT_TIME_INTERVAL)
            payload = {
                "type": "marketsnapshotsubscribe",
            }
            subscribe_orderbook_request: WSJSONRequest = WSJSONRequest(payload=payload)
            await ws.send(subscribe_orderbook_request)
        async for msg in ws.iter_messages():
            data = msg.data
            if data is not None and data["type"] == "marketSnapShot":
                price_list = data["data"]

                last_traded_price = float([i["close"] for i in price_list if i["pair"] == ex_symbol][0])
                payload = {
                    "type": "marketsnapshotunsubscribe",
                }
                subscribe_orderbook_request: WSJSONRequest = WSJSONRequest(payload=payload)
                await ws.send(subscribe_orderbook_request)
                await ws.disconnect()
                return last_traded_price

    async def _make_network_check_request(self):
        await self._api_get(path_url=self.check_network_request_path,
                            headers={"Content-Type": "application/json"},
                            limit_id=CONSTANTS.IP_REQUEST_WEIGHT)

    async def _make_trading_rules_request(self) -> Any:
        exchange_info = await self._api_get(path_url=self.trading_rules_request_path,
                                            headers={"Content-Type": "application/json"},
                                            limit_id=CONSTANTS.IP_REQUEST_WEIGHT)
        return exchange_info

    async def _make_trading_pairs_request(self) -> Any:
        exchange_info = await self._api_get(path_url=self.trading_pairs_request_path,
                                            headers={"Content-Type": "application/json"},
                                            limit_id=CONSTANTS.IP_REQUEST_WEIGHT)
        return exchange_info



import asyncio

from hummingbot.connector.utilities.oms_connector import oms_connector_constants as CONSTANTS
from hummingbot.connector.utilities.oms_connector.oms_connector_auth import OMSConnectorAuth
from hummingbot.connector.utilities.oms_connector.oms_connector_web_utils import (
    OMSConnectorURLCreatorBase,
    OMSConnectorWebAssistantsFactory,
)
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.web_assistant.connections.data_types import WSJSONRequest
from hummingbot.core.web_assistant.ws_assistant import WSAssistant


class OMSConnectorAPIUserStreamDataSource(UserStreamTrackerDataSource):
    def __init__(
        self,
        api_factory: OMSConnectorWebAssistantsFactory,
        url_provider: OMSConnectorURLCreatorBase,
        oms_id: int,
    ):
        super().__init__()
        self._api_factory = api_factory
        self._auth: OMSConnectorAuth = api_factory.auth
        self._url_provider = url_provider
        self._oms_id = oms_id

    async def _connected_websocket_assistant(self) -> WSAssistant:
        ws: WSAssistant = await self._get_ws_assistant()
        url = self._url_provider.get_ws_url()
        await ws.connect(ws_url=url, message_timeout=CONSTANTS.WS_MESSAGE_TIMEOUT)
        auth_payload = {
            CONSTANTS.MSG_ENDPOINT_FIELD: CONSTANTS.WS_AUTH_ENDPOINT,
            CONSTANTS.MSG_DATA_FIELD: {},
        }
        auth_request = WSJSONRequest(
            payload=auth_payload, throttler_limit_id=CONSTANTS.WS_AUTH_ENDPOINT, is_auth_required=True
        )
        await ws.send(auth_request)
        return ws

    async def _get_ws_assistant(self) -> WSAssistant:
        if self._ws_assistant is None:
            self._ensure_authenticated()
            self._ws_assistant = await self._api_factory.get_ws_assistant()
        return self._ws_assistant

    def _ensure_authenticated(self):
        if not self._auth.initialized:
            raise RuntimeError("The authenticator is not initialized.")

    async def _subscribe_channels(self, websocket_assistant: WSAssistant):
        try:
            payload = {
                CONSTANTS.MSG_ENDPOINT_FIELD: CONSTANTS.WS_ACC_EVENTS_ENDPOINT,
                CONSTANTS.MSG_DATA_FIELD: {
                    CONSTANTS.ACCOUNT_ID_FIELD: self._api_factory.auth.account_id,
                    CONSTANTS.OMS_ID_FIELD: self._oms_id,
                },
            }
            subscribe_account_request = WSJSONRequest(payload=payload, is_auth_required=True)

            async with self._api_factory.throttler.execute_task(limit_id=CONSTANTS.WS_REQ_LIMIT_ID):
                await websocket_assistant.send(subscribe_account_request)

            self.logger().info("Subscribed to private account and orders channels...")
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().exception("Unexpected error occurred subscribing to order book trading and delta streams...")
            raise

    async def _process_websocket_messages(self, websocket_assistant: WSAssistant, queue: asyncio.Queue):
        while True:
            try:
                async for ws_response in websocket_assistant.iter_messages():
                    data = ws_response.data
                    if data[CONSTANTS.MSG_TYPE_FIELD] == CONSTANTS.EVENT_MSG_TYPE:
                        await self._process_event_message(event_message=data, queue=queue)
            except asyncio.TimeoutError:
                ping_payload = {
                    CONSTANTS.MSG_ENDPOINT_FIELD: CONSTANTS.WS_PING_REQUEST,
                    CONSTANTS.MSG_DATA_FIELD: {},
                }
                ping_request = WSJSONRequest(payload=ping_payload)
                async with self._api_factory.throttler.execute_task(limit_id=CONSTANTS.WS_PING_REQUEST):
                    await websocket_assistant.send(request=ping_request)



from copy import deepcopy
from typing import Any, Dict, Iterable, List, Optional

import pandas as pd


def flatten(items):
    """
    Deep flatten any iterable item.
    """
    for x in items:
        if isinstance(x, Iterable) and not isinstance(x, (str, bytes)):
            yield from flatten(x)
        else:
            yield x


def list_gateway_wallets(wallets: List[Any], chain: str) -> List[str]:
    """
    Get the public keys for a chain supported by gateway.
    """
    return list(flatten([w["walletAddresses"] for w in wallets if w["chain"] == chain]))


def build_wallet_display(native_token: str, wallets: List[Dict[str, Any]]) -> pd.DataFrame:
    """
    Display user wallets for a particular chain as a table
    """
    columns = ["Wallet", native_token]
    data = []
    for dict in wallets:
        data.extend([[dict['address'], dict['balance']]])

    return pd.DataFrame(data=data, columns=columns)


def build_connector_display(connectors: List[Dict[str, Any]]) -> pd.DataFrame:
    """
    Display connector information as a table
    """
    columns = ["Exchange", "Network", "Wallet"]
    data = []
    for connector_spec in connectors:
        data.extend([
            [
                connector_spec["connector"],
                f"{connector_spec['chain']} - {connector_spec['network']}",
                connector_spec["wallet_address"],
            ]
        ])

    return pd.DataFrame(data=data, columns=columns)


def build_list_display(connectors: List[Dict[str, Any]]) -> pd.DataFrame:
    """
    Display connector information as a table
    """
    columns = ["Exchange", "Chains"]
    data = []
    for connector_spec in connectors:
        data.extend([
            [
                connector_spec["name"],
                ', '.join(connector_spec['chains']),
            ]
        ])

    return pd.DataFrame(data=data, columns=columns)


def build_connector_tokens_display(connectors_chain_network: List[Dict[str, Any]]) -> pd.DataFrame:
    """
    Display connector and the tokens the balance command will report on
    """
    columns = ["Exchange", "Report Token Balances"]
    data = []
    for connector_spec in connectors_chain_network:
        data.extend([
            [
                f"{connector_spec['connector']}_{connector_spec['chain']}_{connector_spec['network']}",
                connector_spec.get("tokens", ""),
            ]
        ])

    return pd.DataFrame(data=data, columns=columns)


def build_balances_allowances_display(symbols: List[str], balances: List[str], allowances: List[str]) -> pd.DataFrame:
    """
    Display balances and allowances for a list of symbols as a table
    """
    columns = ["Symbol", "Balance", "Allowances"]
    data = []
    for i in range(len(symbols)):
        data.extend([
            [
                symbols[i],
                balances[i],
                allowances[i]
            ]
        ])

    return pd.DataFrame(data=data, columns=columns)


def build_config_dict_display(lines: List[str], config_dict: Dict[str, Any], level: int = 0):
    """
    Build display messages on lines for a config dictionary, this function is called recursive.
    For example:
    config_dict: {"a": 1, "b": {"ba": 2, "bb": 3}, "c": 4}
    lines will be
    a: 1
    b:
      ba: 2
      bb: 3
    c: 4
    :param lines: a list display message (lines) to be built upon.
    :param config_dict: a (Gateway) config dictionary
    :param level: a nested level.
    """
    prefix: str = "  " * level
    for k, v in config_dict.items():
        if isinstance(v, Dict):
            lines.append(f"{prefix}{k}:")
            build_config_dict_display(lines, v, level + 1)
        else:
            lines.append(f"{prefix}{k}: {v}")


def build_config_namespace_keys(namespace_keys: List[str], config_dict: Dict[str, Any], prefix: str = ""):
    """
    Build namespace keys for a config dictionary, this function is recursive.
    For example:
    config_dict: {"a": 1, "b": {"ba": 2, "bb": 3}, "c": 4}
    namepace_keys will be ["a", "b", "b.ba", "b.bb", "c"]
    :param namespace_keys: a key list to be build upon
    :param config_dict: a (Gateway) config dictionary
    :prefix: a prefix to the namespace (used when the function is called recursively.
    """
    for k, v in config_dict.items():
        namespace_keys.append(f"{prefix}{k}")
        if isinstance(v, Dict):
            build_config_namespace_keys(namespace_keys, v, f"{prefix}{k}.")


def search_configs(config_dict: Dict[str, Any], namespace_key: str) \
        -> Optional[Dict[str, Any]]:
    """
    Search the config dictionary for a given namespace key and preserve the key hierarchy.
    For example:
    config_dict:  {"a": 1, "b": {"ba": 2, "bb": 3}, "c": 4}
    searching for b will result in {"b": {"ba": 2, "bb": 3}}
    searching for b.ba will result in {"b": {"ba": 2}}
    :param config_dict: The config dictionary
    :param namespace_key: The namespace key to search for
    :return: A dictionary matching the given key, returns None if not found
    """
    key_parts = namespace_key.split(".")
    if not key_parts[0] in config_dict:
        return
    result: Dict[str, Any] = {key_parts[0]: deepcopy(config_dict[key_parts[0]])}
    result_val = result[key_parts[0]]
    for key_part in key_parts[1:]:
        if not isinstance(result_val, Dict) or key_part not in result_val:
            return
        temp = deepcopy(result_val[key_part])
        result_val.clear()
        result_val[key_part] = temp
        result_val = result_val[key_part]
    return result



import numpy as np
import logging
cimport numpy as np


pmm_logger = None

cdef class RingBuffer:
    @classmethod
    def logger(cls):
        global pmm_logger
        if pmm_logger is None:
            pmm_logger = logging.getLogger(__name__)
        return pmm_logger

    def __cinit__(self, int length):
        self._length = length
        self._buffer = np.zeros(length, dtype=np.float64)
        self._delimiter = 0
        self._is_full = False

    def __dealloc__(self):
        self._buffer = None

    cdef void c_add_value(self, float val):
        self._buffer[self._delimiter] = val
        self.c_increment_delimiter()

    cdef void c_increment_delimiter(self):
        self._delimiter = (self._delimiter + 1) % self._length
        if not self._is_full and self._delimiter == 0:
            self._is_full = True

    cdef bint c_is_empty(self):
        return (not self._is_full) and (0==self._delimiter)

    cdef double c_get_last_value(self):
        if self.c_is_empty():
            return np.nan
        return self._buffer[self._delimiter-1]

    cdef bint c_is_full(self):
        return self._is_full

    cdef double c_mean_value(self):
        result = np.nan
        if self._is_full:
            result=np.mean(self.c_get_as_numpy_array())
        return result

    cdef double c_variance(self):
        result = np.nan
        if self._is_full:
            result = np.var(self.c_get_as_numpy_array())
        return result

    cdef double c_std_dev(self):
        result = np.nan
        if self._is_full:
            result = np.std(self.c_get_as_numpy_array())
        return result

    cdef np.ndarray[np.double_t, ndim=1] c_get_as_numpy_array(self):
        cdef np.ndarray[np.int16_t, ndim=1] indexes

        if not self._is_full:
            indexes = np.arange(0, stop=self._delimiter, dtype=np.int16)
        else:
            indexes = np.arange(self._delimiter, stop=self._delimiter + self._length,
                                dtype=np.int16) % self._length
        return np.asarray(self._buffer)[indexes]

    def __init__(self, length):
        self._length = length
        self._buffer = np.zeros(length, dtype=np.double)
        self._delimiter = 0
        self._is_full = False

    def add_value(self, val):
        self.c_add_value(val)

    def get_as_numpy_array(self):
        return self.c_get_as_numpy_array()

    def get_last_value(self):
        return self.c_get_last_value()

    @property
    def is_full(self):
        return self.c_is_full()

    @property
    def mean_value(self):
        return self.c_mean_value()

    @property
    def std_dev(self):
        return self.c_std_dev()

    @property
    def variance(self):
        return self.c_variance()

    @property
    def length(self) -> int:
        return self._length

    @length.setter
    def length(self, value):
        data = self.get_as_numpy_array()

        self._length = value
        self._buffer = np.zeros(value, dtype=np.float64)
        self._delimiter = 0
        self._is_full = False

        for val in data[-value:]:
            self.add_value(val)



import asyncio
from decimal import Decimal
from typing import Any, AsyncIterable, Dict, List, Optional

from bidict import bidict

import hummingbot.connector.exchange.htx.htx_constants as CONSTANTS
from hummingbot.connector.constants import s_decimal_0, s_decimal_NaN
from hummingbot.connector.exchange.htx import htx_web_utils as web_utils
from hummingbot.connector.exchange.htx.htx_api_order_book_data_source import HtxAPIOrderBookDataSource
from hummingbot.connector.exchange.htx.htx_api_user_stream_data_source import HtxAPIUserStreamDataSource
from hummingbot.connector.exchange.htx.htx_auth import HtxAuth
from hummingbot.connector.exchange.htx.htx_utils import is_exchange_information_valid
from hummingbot.connector.exchange_py_base import ExchangePyBase
from hummingbot.connector.trading_rule import TradingRule
from hummingbot.connector.utils import combine_to_hb_trading_pair
from hummingbot.core.data_type.common import OrderType, TradeType
from hummingbot.core.data_type.in_flight_order import InFlightOrder, OrderUpdate, TradeUpdate
from hummingbot.core.data_type.trade_fee import TokenAmount, TradeFeeBase
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.core.utils.estimate_fee import build_trade_fee
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory


class HtxExchange(ExchangePyBase):

    web_utils = web_utils

    def __init__(
        self,
        htx_api_key: str,
        htx_secret_key: str,
        balance_asset_limit: Optional[Dict[str, Dict[str, Decimal]]] = None,
        rate_limits_share_pct: Decimal = Decimal("100"),
        trading_pairs: Optional[List[str]] = None,
        trading_required: bool = True,
    ):
        self.htx_api_key = htx_api_key
        self.htx_secret_key = htx_secret_key
        self._trading_pairs = trading_pairs
        self._trading_required = trading_required
        self._account_id = ""
        super().__init__(balance_asset_limit, rate_limits_share_pct)

    @property
    def name(self) -> str:
        return "htx"

    @property
    def authenticator(self):
        return HtxAuth(
            api_key=self.htx_api_key, secret_key=self.htx_secret_key, time_provider=self._time_synchronizer
        )

    @property
    def rate_limits_rules(self):
        return CONSTANTS.RATE_LIMITS

    @property
    def domain(self):
        return CONSTANTS.DOMAIN

    @property
    def client_order_id_max_length(self):
        return CONSTANTS.MAX_CLIENT_ORDER_ID_LENGTH

    @property
    def client_order_id_prefix(self):
        return CONSTANTS.BROKER_ID

    @property
    def trading_rules_request_path(self):
        return CONSTANTS.TRADE_INFO_URL

    @property
    def trading_pairs_request_path(self):
        return CONSTANTS.TRADE_INFO_URL

    @property
    def check_network_request_path(self):
        return CONSTANTS.SERVER_TIME_URL

    @property
    def trading_pairs(self):
        return self._trading_pairs

    @property
    def is_cancel_request_in_exchange_synchronous(self) -> bool:
        return False

    @property
    def is_trading_required(self) -> bool:
        return self._trading_required

    def supported_order_types(self):
        return [OrderType.LIMIT, OrderType.LIMIT_MAKER, OrderType.MARKET]

    def get_fee(
        self,
        base_currency: str,
        quote_currency: str,
        order_type: OrderType,
        order_side: TradeType,
        amount: Decimal,
        price: Decimal = s_decimal_NaN,
        is_maker: Optional[bool] = None,
    ):
        return build_trade_fee(
            self.name,
            is_maker,
            base_currency=base_currency,
            quote_currency=quote_currency,
            order_type=order_type,
            order_side=order_side,
            amount=amount,
            price=price,
        )

    def _is_request_exception_related_to_time_synchronizer(self, request_exception: Exception):
        # API documentation does not clarify the error message for timestamp related problems
        return False

    def _is_order_not_found_during_status_update_error(self, status_update_exception: Exception) -> bool:
        # TODO: implement this method correctly for the connector
        # The default implementation was added when the functionality to detect not found orders was introduced in the
        # ExchangePyBase class. Also fix the unit test test_lost_order_removed_if_not_found_during_order_status_update
        # when replacing the dummy implementation
        return False

    def _is_order_not_found_during_cancelation_error(self, cancelation_exception: Exception) -> bool:
        # TODO: implement this method correctly for the connector
        # The default implementation was added when the functionality to detect not found orders was introduced in the
        # ExchangePyBase class. Also fix the unit test test_cancel_order_not_found_in_the_exchange when replacing the
        # dummy implementation
        return False

    def _create_web_assistants_factory(self) -> WebAssistantsFactory:
        return web_utils.build_api_factory(
            throttler=self._throttler, time_synchronizer=self._time_synchronizer, auth=self._auth
        )

    def _create_order_book_data_source(self):
        return HtxAPIOrderBookDataSource(
            trading_pairs=self.trading_pairs, connector=self, api_factory=self._web_assistants_factory
        )

    def _create_user_stream_data_source(self) -> UserStreamTrackerDataSource:
        return HtxAPIUserStreamDataSource(
            htx_auth=self._auth,
            trading_pairs=self._trading_pairs,
            connector=self,
            api_factory=self._web_assistants_factory,
        )

    def _get_fee(
        self,
        base_currency: str,
        quote_currency: str,
        order_type: OrderType,
        order_side: TradeType,
        amount: Decimal,
        price: Decimal = s_decimal_NaN,
        is_maker: Optional[bool] = None,
    ) -> TradeFeeBase:

        is_maker = is_maker or (order_type is OrderType.LIMIT_MAKER)
        fee = build_trade_fee(
            self.name,
            is_maker,
            base_currency=base_currency,
            quote_currency=quote_currency,
            order_type=order_type,
            order_side=order_side,
            amount=amount,
            price=price,
        )
        return fee

    async def _update_account_id(self) -> str:
        accounts = await self._api_get(path_url=CONSTANTS.ACCOUNT_ID_URL, is_auth_required=True)
        try:
            for account in accounts["data"]:
                if account["state"] == "working" and account["type"] == "spot":
                    self._account_id = str(account["id"])
        except Exception:
            raise ValueError(f"Unable to retrieve account id.\n{accounts['err-msg']}")

    async def _update_balances(self):

        new_available_balances = {}
        new_balances = {}
        if not self._account_id:
            await self._update_account_id()
        data = await self._api_get(
            path_url=CONSTANTS.ACCOUNT_BALANCE_URL.format(self._account_id),
            is_auth_required=True,
            limit_id=CONSTANTS.ACCOUNT_BALANCE_LIMIT_ID,
        )
        balances = data.get("data", {}).get("list", [])
        if len(balances) > 0:
            for balance_entry in balances:
                asset_name = balance_entry["currency"].upper()
                balance = Decimal(balance_entry["balance"])
                if balance == s_decimal_0:
                    continue
                if asset_name not in new_available_balances:
                    new_available_balances[asset_name] = s_decimal_0
                if asset_name not in new_balances:
                    new_balances[asset_name] = s_decimal_0

                new_balances[asset_name] += balance
                if balance_entry["type"] == "trade":
                    new_available_balances[asset_name] = balance

            self._account_available_balances = new_available_balances
            self._account_balances = new_balances

    async def _format_trading_rules(self, raw_trading_pair_info: List[Dict[str, Any]]) -> List[TradingRule]:
        trading_rules = []
        supported_symbols = await self.trading_pair_symbol_map()
        for info in raw_trading_pair_info["data"]:
            try:
                if info["symbol"] not in supported_symbols:
                    continue
                base_asset = info["bc"]
                quote_asset = info["qc"]
                price_precision = info["pp"]
                amount_precision = info["ap"]
                value_precision = info["vp"]
                trading_rules.append(
                    TradingRule(
                        trading_pair=f"{base_asset}-{quote_asset}".upper(),
                        min_order_size=Decimal(info["minoa"]),
                        max_order_size=Decimal(info["maxoa"]),
                        min_price_increment=Decimal(str(10**-price_precision)),
                        min_base_amount_increment=Decimal(str(10**-amount_precision)),
                        min_quote_amount_increment=Decimal(str(10**-value_precision)),
                        min_notional_size=Decimal(info["minov"]),
                    )
                )
            except Exception:
                self.logger().error(f"Error parsing the trading pair rule {info}. Skipping.", exc_info=True)
        return trading_rules

    async def _all_trade_updates_for_order(self, order: InFlightOrder) -> List[TradeUpdate]:
        trade_updates = []

        if order.exchange_order_id is not None:
            exchange_order_id = int(order.exchange_order_id)
            all_fills_response = await self._api_get(
                path_url=CONSTANTS.ORDER_MATCHES_URL.format(exchange_order_id),
                is_auth_required=True,
                limit_id=CONSTANTS.ORDER_MATCHES_LIMIT_ID,
            )

            for trade in all_fills_response.get("data", []):
                fee = TradeFeeBase.new_spot_fee(
                    fee_schema=self.trade_fee_schema(),
                    trade_type=order.trade_type,
                    percent_token=trade["fee-currency"].upper(),
                    flat_fees=[TokenAmount(amount=Decimal(trade["filled-fees"]), token=trade["fee-currency"].upper())],
                )
                trade_update = TradeUpdate(
                    trade_id=str(trade["trade-id"]),
                    client_order_id=order.client_order_id,
                    exchange_order_id=str(trade["order-id"]),
                    trading_pair=order.trading_pair,
                    fee=fee,
                    fill_base_amount=Decimal(trade["filled-amount"]),
                    fill_quote_amount=Decimal(trade["filled-amount"]) * Decimal(trade["price"]),
                    fill_price=Decimal(trade["price"]),
                    fill_timestamp=trade["created-at"] * 1e-3,
                )
                trade_updates.append(trade_update)

        return trade_updates

    async def _request_order_status(self, tracked_order: InFlightOrder) -> OrderUpdate:
        exchange_order_id = await tracked_order.get_exchange_order_id()
        updated_order_data = await self._api_get(
            path_url=CONSTANTS.ORDER_DETAIL_URL.format(exchange_order_id),
            is_auth_required=True,
            limit_id=CONSTANTS.ORDER_DETAIL_LIMIT_ID,
        )

        if updated_order_data["status"] == "ok":
            new_state = CONSTANTS.ORDER_STATE[updated_order_data["data"]["state"]]

            order_update = OrderUpdate(
                client_order_id=tracked_order.client_order_id,
                exchange_order_id=exchange_order_id,
                trading_pair=tracked_order.trading_pair,
                update_timestamp=self.current_timestamp,
                new_state=new_state,
            )

            return order_update
        else:
            raise ValueError(f"Erroneous order status response {updated_order_data}")

    async def _iter_user_event_queue(self) -> AsyncIterable[Dict[str, Any]]:
        """
        Called by _user_stream_event_listener.
        """
        while True:
            try:
                yield await self._user_stream_tracker.user_stream.get()
            except asyncio.CancelledError:
                raise
            except Exception as e:
                self.logger().error(f"Unknown error. Retrying after 1 second. {e}", exc_info=True)
                await asyncio.sleep(1.0)

    async def _user_stream_event_listener(self):
        """
        This functions runs in background continuously processing the events received from the exchange by the user
        stream data source. It keeps reading events from the queue until the task is interrupted.
        The events received are balance updates, order updates and trade events.
        """
        async for stream_message in self._iter_user_event_queue():
            try:
                channel = stream_message["ch"]
                data = stream_message["data"]
                if channel.startswith("accounts"):
                    asset_name = data["currency"].upper()
                    balance = data["balance"]
                    available_balance = data["available"]

                    self._account_balances.update({asset_name: Decimal(balance)})
                    self._account_available_balances.update({asset_name: Decimal(available_balance)})
                elif channel.startswith("orders"):
                    safe_ensure_future(self._process_order_update(data))
                elif channel.startswith("trade.clearing"):
                    safe_ensure_future(self._process_trade_event(data))

            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().error("Unexpected error in user stream listener loop.", exc_info=True)
                await self._sleep(5.0)

    async def _process_order_update(self, msg: Dict[str, Any]):
        client_order_id = msg["clientOrderId"]
        order_status = msg["orderStatus"]
        tracked_order = self._order_tracker.all_updatable_orders.get(client_order_id)
        if tracked_order is not None:
            order_update = OrderUpdate(
                trading_pair=tracked_order.trading_pair,
                update_timestamp=self.current_timestamp,
                new_state=CONSTANTS.ORDER_STATE[order_status],
                client_order_id=client_order_id,
            )
            self._order_tracker.process_order_update(order_update=order_update)

    async def _process_trade_event(self, trade_event: Dict[str, Any]):
        client_order_id = trade_event["clientOrderId"]
        tracked_order = self._order_tracker.all_fillable_orders.get(client_order_id)

        if tracked_order:
            fee = TradeFeeBase.new_spot_fee(
                fee_schema=self.trade_fee_schema(),
                trade_type=tracked_order.trade_type,
                percent_token=trade_event["feeCurrency"].upper(),
                flat_fees=[
                    TokenAmount(amount=Decimal(trade_event["transactFee"]), token=trade_event["feeCurrency"].upper())
                ],
            )
            trade_update = TradeUpdate(
                trade_id=str(trade_event["tradeId"]),
                client_order_id=client_order_id,
                exchange_order_id=str(trade_event["orderId"]),
                trading_pair=tracked_order.trading_pair,
                fee=fee,
                fill_base_amount=Decimal(trade_event["tradeVolume"]),
                fill_quote_amount=Decimal(trade_event["tradeVolume"]) * Decimal(trade_event["tradePrice"]),
                fill_price=Decimal(trade_event["tradePrice"]),
                fill_timestamp=trade_event["tradeTime"] * 1e-3,
            )
            self._order_tracker.process_trade_update(trade_update)

    async def _update_trading_fees(self):
        pass

    async def _place_order(
        self,
        order_id: str,
        trading_pair: str,
        amount: Decimal,
        trade_type: TradeType,
        order_type: OrderType,
        price: Decimal,
        **kwargs,
    ):
        path_url = CONSTANTS.PLACE_ORDER_URL
        side = trade_type.name.lower()
        if order_type.is_limit_type():
            order_type_str = "limit" if order_type is OrderType.LIMIT else "limit-maker"
        else:
            order_type_str = "market"
        if not self._account_id:
            await self._update_account_id()
        exchange_symbol = await self.exchange_symbol_associated_to_pair(trading_pair)
        params = {
            "account-id": self._account_id,
            "amount": f"{amount}",
            "client-order-id": order_id,
            "symbol": exchange_symbol,
            "type": f"{side}-{order_type_str}",
        }
        if order_type is OrderType.LIMIT or order_type is OrderType.LIMIT_MAKER:
            params["price"] = f"{price}"
        creation_response = await self._api_post(path_url=path_url, params=params, data=params, is_auth_required=True)

        if (
            creation_response["status"] == "ok"
            and creation_response["data"] is not None
            and str(creation_response["data"]).isdecimal()
        ):
            exchange_order_id = str(creation_response["data"])
            return exchange_order_id, self.current_timestamp
        else:
            raise ValueError(f"Htx rejected the order {order_id} ({creation_response})")

    async def _place_cancel(self, order_id: str, tracked_order: InFlightOrder):
        if tracked_order is None:
            raise ValueError(f"Failed to cancel order - {order_id}. Order not found.")
        path_url = CONSTANTS.CANCEL_ORDER_URL.format(tracked_order.exchange_order_id)
        params = {"order-id": str(tracked_order.exchange_order_id)}
        response = await self._api_post(
            path_url=path_url, params=params, data=params, limit_id=CONSTANTS.CANCEL_URL_LIMIT_ID, is_auth_required=True
        )
        if response.get("status") == "ok":
            return True
        return False

    def _initialize_trading_pair_symbols_from_exchange_info(self, exchange_info: Dict[str, Any]):
        mapping = bidict()
        for symbol_data in filter(is_exchange_information_valid, exchange_info.get("data", [])):
            mapping[symbol_data["symbol"]] = combine_to_hb_trading_pair(
                base=symbol_data["bc"].upper(), quote=symbol_data["qc"].upper()
            )

        self._set_trading_pair_symbol_map(mapping)

    async def _get_last_traded_price(self, trading_pair: str) -> float:
        path_url = CONSTANTS.MOST_RECENT_TRADE_URL
        params = {"symbol": await self.exchange_symbol_associated_to_pair(trading_pair)}
        resp_json = await self._api_get(
            path_url=path_url,
            params=params,
        )
        resp_record = resp_json["tick"]["data"][0]
        return float(resp_record["price"])



import hashlib
import hmac
import time
from typing import Any, Dict

from hummingbot.connector.exchange.ascend_ex import ascend_ex_constants as CONSTANTS
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.data_types import RESTRequest, WSRequest


class AscendExAuth(AuthBase):
    """
    Auth class required by AscendEx API
    Learn more at https://ascendex.github.io/ascendex-pro-api/#authenticate-a-restful-request
    """

    def __init__(self, api_key: str, secret_key: str):
        self.api_key = api_key
        self.secret_key = secret_key

    async def rest_authenticate(self, request: RESTRequest) -> RESTRequest:
        """
        Adds the server time and the signature to the request, required for authenticated interactions. It also adds
        the required parameter in the request header.
        :param request: the request to be configured for authenticated interaction
        """
        # Generates auth headers
        path = request.throttler_limit_id
        if path != CONSTANTS.BALANCE_HISTORY_PATH_URL:
            path = path.replace("cash/", "").replace("spot/", "")
        headers_auth = self.get_auth_headers(path)

        headers = {}
        if request.headers is not None:
            headers.update(request.headers)
        headers.update(headers_auth)
        request.headers = headers

        return request

    async def ws_authenticate(self, request: WSRequest) -> WSRequest:
        """
        This method is intended to configure a websocket request to be authenticated. AscendEx does not use this
        functionality
        """
        return request  # pass-through

    def get_auth_headers(self, path_url: str, data: Dict[str, Any] = None):
        """
        Generates authentication signature and return it in a dictionary along with other inputs
        :param path_url: URL of the auth API endpoint
        :param data: data to be included in the headers
        :return: a dictionary of request info including the request signature
        """

        timestamp = str(int(self._time() * 1e3))
        message = timestamp + path_url
        signature = hmac.new(self.secret_key.encode("utf-8"), message.encode("utf-8"), hashlib.sha256).hexdigest()

        return {
            "x-auth-key": self.api_key,
            "x-auth-signature": signature,
            "x-auth-timestamp": timestamp,
        }

    def _time(self) -> float:
        return time.time()
